from handwriting_features.interface.featurizer.handlers import MultiSubjectFeatureExtractorHandler


class FeatureExtractor(object):
    """
    Class implementing the features extractor interface for the Featurizer API.

    For more information about featurizer, see the following repositories:
    1. [server side](#github.com/BDALab/featurizer-api)
    1. [client side](#github.com/BDALab/featurizer-api-client)

    For more information about the attributes, see: ``extract(...)``
    """

    def __init__(self, values, labels=None, **configuration):
        """
        Initializes the FeatureExtractor featurizer API interface.

        :param values: data values to extract the features from
        :type values: numpy.ndarray
        :param labels: data labels for data samples, defaults to None
        :type labels: list, optional
        :param configuration: common extractor configuration
        :type configuration: **kwargs, optional
        """

        # Set the sample values/labels
        self.values = values
        self.labels = labels if labels else []

        # Set the extractor configuration
        self.configuration = configuration if configuration else {}

        # Initialize the handler
        self.handler = MultiSubjectFeatureExtractorHandler

    def extract(self, pipeline):
        """
        Interface method: extract the features.

        **Data**

        1. data is of type: ``numpy.ndarray``.
        2. data is mandatory.
        3. data shape: In general, data to have the shape (M, ..., D). Where M
           stands for subjects (i.e. subjects are in the first dimension), and
           D stands for D data samples (of shape ...).
            1. in the case of data having the following shape: (D, ), the API
               assumes it is a vector of D data sample points for one subject.
               It transforms the data to a row vector: (1, D) to add the
               dimension for the subject.
            2: in the case of data having the following shape: (M, ..., D),
               the API does not transform the data, but it assumes there are
               M subjects abd D data samples, each having (...) dimensionality,
               e.g. if data has the shape (M, 3, 10) it means that there are
               M subjects and each of the subjects has 10 data samples (each
               being three dimensional).

        **Labels**

        1. labels are of type: ``list``.
        2. labels are optional.
        3. labels are of length D (for each data sample, there is one label)

        **Configuration**

        1. configuration are of type: ``dict``.
        2. configuration is optional.
        3. configuration provides common kwargs for feature extraction

        **Pipeline**

        1. pipeline is of type: ``list``.
        2. pipeline is mandatory.
        3. each element in the pipeline is of type: ``dict``.
        4. each element in the pipeline has the following keys: a) ``name``
           to hold the name of the feature to be computed, and b) ``args``
           to hold the arguments (kwargs) for the specific feature extraction
           method that is going to be used (it is of type: ``dict``).

        **Output**

        The extracted features follow the same shape convention as the input
        data: the subjects are in the first dimension, and the features are
        in the last dimension (each feature having shape ...).

        :param pipeline: pipeline of the features to be extracted
        :type pipeline: list
        :return: extracted features and labels
        :rtype: dict {"features": ..., "labels": ...}
        """
        return self.handler.extract(self.values, self.labels, pipeline, **self.configuration)



import numpy
import functools
from copy import deepcopy
from handwriting_features.data.utils.math import intersection
from handwriting_features.data.utils.math import derivation
from handwriting_features.data.utils.dsp import GaussianFilter, segment


class IntersectionUtils(object):
    """Class implementing intersection utils"""

    def __init__(self, sample_wrapper):
        """Initializes writing intersection object"""

        # Set the sample wrapper
        self.sample_wrapper = deepcopy(sample_wrapper)

        # Get the duration
        self.duration = self.sample_wrapper.sample_time[-1] - self.sample_wrapper.sample_time[0]

        # Get the on-surface strokes
        self.on_surface_strokes = self.sample_wrapper.on_surface_strokes

        # Set the number of intra-stroke and inter-stroke intersections per stroke
        self.abs_num_intra = numpy.zeros((len(self.on_surface_strokes), 1), dtype=int)
        self.rel_num_intra = numpy.zeros((len(self.on_surface_strokes), 1), dtype=float)
        self.abs_num_inter = 0
        self.rel_num_inter = 0.

        # Set the concatenated vector of strokes and intersections
        self.all_x = []
        self.all_y = []
        self.intra_all = []

        # Compute the intra-stroke and inter-stroke intersections
        self._compute_intersections()

    def get_number_of_intra_stroke_intersections(self):
        """Extracts the number of intra-stroke intersections"""
        return numpy.NaN if not self.on_surface_strokes else self.abs_num_intra.T

    def get_relative_number_of_intra_stroke_intersections(self):
        """Extracts the relative number of intra-stroke intersections"""
        return numpy.NaN if not self.on_surface_strokes else self.rel_num_intra.T

    def get_total_number_of_intra_stroke_intersections(self):
        """Extracts the total number of intra-stroke intersections"""
        return 0 if not self.on_surface_strokes else int(numpy.sum(self.abs_num_intra))

    def get_relative_total_number_of_intra_stroke_intersections(self):
        """Extracts the relative total number of intra-stroke intersections"""
        return 0. if not self.on_surface_strokes else round(float(numpy.sum(self.abs_num_intra) / self.duration), 6)

    def get_number_of_inter_stroke_intersections(self):
        """Extracts the number of inter-stroke intersections"""
        return 0 if not self.on_surface_strokes else self.abs_num_inter

    def get_relative_number_of_inter_stroke_intersections(self):
        """Extracts the relative number of inter-stroke intersections"""
        return 0. if not self.on_surface_strokes else round(float(self.rel_num_inter), 6)

    @functools.lru_cache(maxsize=1)
    def _compute_intersections(self):
        """Computes the intra-stroke and inter-stroke intersections"""

        # Compute the intra-stroke intersections
        for i, stroke in enumerate(self.on_surface_strokes):
            intersections = [(x, y) for x, y in intersection(stroke.x, stroke.y) if x is not None and y is not None]

            # Handle the intra-stroke intersections
            if intersections:
                self.abs_num_intra[i] = len(intersections)
                self.rel_num_intra[i] = len(intersections) / (stroke.time[-1] - stroke.time[0])
                self.intra_all.extend(intersections)

            # Add the stroke to the concatenated vector
            self.all_x.extend([x for x in stroke.x] + [numpy.nan])
            self.all_y.extend([y for y in stroke.y] + [numpy.nan])

        self.all_x = numpy.array(self.all_x)
        self.all_y = numpy.array(self.all_y)

        # Compute the inter-stroke intersections
        intersections = [(x, y) for x, y in intersection(self.all_x, self.all_y) if x is not None and y is not None]

        # Handle the inter-stroke intersections
        if self.intra_all:
            intersections = [i for i in intersections if i not in self.intra_all]
        if intersections:
            self.abs_num_inter = len(intersections)
            self.rel_num_inter = len(intersections) / self.duration
        else:
            self.abs_num_inter = 0
            self.rel_num_inter = 0.


class ProjectionUtils(object):
    """Class implementing projection utils"""

    # Default computational arguments
    n = 50

    def __init__(self, sample_wrapper, fs, n=None):
        """
        Initializes writing projection object.

        :param fs: sampling frequency
        :type fs: float
        :param n: number of samples of a Gaussian filter, defaults to 50
        :type n: int
        """

        # Set the sample wrapper
        self.sample_wrapper = deepcopy(sample_wrapper)

        # Set the on-surface data
        self.on_surface_data = self.sample_wrapper.on_surface_data

        # Set the DSP arguments
        self.fs = fs
        self.n = n if n else ProjectionUtils.n

        # Set the properties of the windowing function
        self.window_size = int(numpy.ceil(self.fs * 0.100))
        self.window_step = self.window_size

        # Set the filter instances
        self.gaussian_filter = GaussianFilter(self.fs, self.n)

        # Prepare time, peaks and valleys
        self.time = None
        self.peaks = None
        self.valleys = None

        # Compute the projections
        self._compute_projections()

    @functools.lru_cache(maxsize=1)
    def _compute_projections(self):
        """Computes the projections"""

        # Take time and vertical movement only
        t = self.on_surface_data.time
        y = self.on_surface_data.y

        # Segment time and vertical movement
        t_segmented = segment(t, self.window_size, self.window_step)
        y_segmented = segment(y, self.window_size, self.window_step)

        # Reshape the segments
        t_segmented = numpy.reshape(t_segmented, (t_segmented.shape[0], t_segmented.shape[-1])).T
        y_segmented = numpy.reshape(y_segmented, (y_segmented.shape[0], y_segmented.shape[-1])).T

        # Filter out segments with CV < 0.0001
        selection = (numpy.std(y_segmented, axis=0) / numpy.mean(y_segmented, axis=0)) > 0.0001
        t_segmented = t_segmented[:, selection]
        y_segmented = y_segmented[:, selection]

        # Flatten the segments to vectors
        t = t_segmented.flatten("F")
        y = y_segmented.flatten("F")

        # Filter vertical movement
        if len(y) > 3 * (self.n - 1):
            y_filtered = self.gaussian_filter.filter(y)
        else:
            y_filtered = numpy.array((y.tolist() + ([y[-1]] * len(y)))[:len(y)])

        # Compute the peaks and valleys
        peaks = numpy.array((y_filtered[1:-2] > y_filtered[0:-3]) & (y_filtered[1:-2] > y_filtered[2:-1]))
        valleys = numpy.array((y_filtered[1:-2] < y_filtered[0:-3]) & (y_filtered[1:-2] < y_filtered[2:-1]))

        # Set the time, peaks and valleys
        self.time = t
        self.peaks = numpy.array([i + 1 for i, p in enumerate(peaks) if p])
        self.valleys = numpy.array([i + 1 for i, p in enumerate(valleys) if p])

        # Adjust the peaks
        for i, peak in enumerate(self.peaks):
            while y[self.peaks[i]] < y[self.peaks[i] - 1] and y[self.peaks[i]] < y[self.peaks[i] + 1]:
                if y[self.peaks[i] - 1] > y[self.peaks[i] + 1]:
                    self.peaks[i] -= 1
                else:
                    self.peaks[i] += 1

        # Adjust the valleys
        for i, peak in enumerate(self.valleys):
            while y[self.valleys[i]] > y[self.valleys[i] - 1] and y[self.valleys[i]] > y[self.valleys[i] + 1]:
                if y[self.valleys[i] - 1] < y[self.valleys[i] + 1]:
                    self.valleys[i] -= 1
                else:
                    self.valleys[i] += 1

    @functools.lru_cache(maxsize=1)
    def _compute_temporal_velocity(self):
        """Computes the temporal velocity"""

        # Prepare the trajectory and time
        d = numpy.sqrt(derivation(self.on_surface_data.x) ** 2 + derivation(self.on_surface_data.y) ** 2)
        t = derivation(self.on_surface_data.time)

        # Return velocity
        return d / t

    @functools.lru_cache(maxsize=1)
    def vertical_peaks_indices(self):
        """Extracts the vertical peaks indices"""
        if all((self.time is not None, self.peaks is not None)):
            return numpy.array([numpy.where(self.on_surface_data.time == e)[0][0] for e in self.time[self.peaks]])
        else:
            return numpy.nan

    @functools.lru_cache(maxsize=1)
    def vertical_valleys_indices(self):
        """Extracts the vertical valleys indices"""
        if all((self.time is not None, self.valleys is not None)):
            return numpy.array([numpy.where(self.on_surface_data.time == e)[0][0] for e in self.time[self.valleys]])
        else:
            return numpy.nan

    def vertical_peaks_values(self):
        """Extracts the vertical peaks values"""
        if self.vertical_peaks_indices() is not numpy.nan:
            return numpy.array([self.on_surface_data.y[e] for e in self.vertical_peaks_indices()])
        else:
            return numpy.nan

    def vertical_valleys_values(self):
        """Extracts the vertical valleys values"""
        if self.vertical_valleys_indices() is not numpy.nan:
            return numpy.array([self.on_surface_data.y[e] for e in self.vertical_valleys_indices()])
        else:
            return numpy.nan

    def vertical_peaks_velocity(self):
        """Extracts the vertical peaks velocity"""
        if self.vertical_peaks_indices() is not numpy.nan:
            return numpy.array([self._compute_temporal_velocity()[e] for e in self.vertical_peaks_indices()])
        else:
            return numpy.nan

    def vertical_valleys_velocity(self):
        """Extracts the vertical valleys velocity"""
        if self.vertical_valleys_indices() is not numpy.nan:
            return numpy.array([self._compute_temporal_velocity()[e] for e in self.vertical_valleys_indices()])
        else:
            return numpy.nan

    def vertical_peaks_distance(self):
        """Extracts the vertical peaks distance"""
        if self.vertical_peaks_indices() is not numpy.nan and len(self.vertical_peaks_indices()) > 1:
            return derivation(numpy.array([self.on_surface_data.x[e] for e in self.vertical_peaks_indices()]))
        else:
            return numpy.nan

    def vertical_valleys_distance(self):
        """Extracts the vertical valleys distance"""
        if self.vertical_valleys_indices() is not numpy.nan and len(self.vertical_valleys_indices()) > 1:
            return derivation(numpy.array([self.on_surface_data.x[e] for e in self.vertical_valleys_indices()]))
        else:
            return numpy.nan

    def vertical_peaks_duration(self):
        """Extracts the vertical peaks duration"""
        if self.vertical_peaks_indices() is not numpy.nan and len(self.vertical_peaks_indices()) > 1:
            return derivation(numpy.array([self.on_surface_data.time[e] for e in self.vertical_peaks_indices()]))
        else:
            return numpy.nan

    def vertical_valleys_duration(self):
        """Extracts the vertical valleys duration"""
        if self.vertical_valleys_indices() is not numpy.nan and len(self.vertical_valleys_indices()) > 1:
            return derivation(numpy.array([self.on_surface_data.time[e] for e in self.vertical_valleys_indices()]))
        else:
            return numpy.nan



import sys
import logging
import numpy
from handwriting_features.features.configuration.settings import HandwritingFeaturesSettings
from handwriting_features.data.descriptors.statistics import Statistics
from handwriting_features.data.containers.sample import HandwritingSampleWrapper
from handwriting_features.features.validation import HandwritingFeaturesFusion, HandwritingFeaturesValidation


# Set the logging configuration
logging.basicConfig(stream=sys.stdout, level=logging.DEBUG)
logger = logging.getLogger("FeatureExtractor")


class HandwritingFeaturesBase(object):
    """Base class for handwriting features"""

    def __init__(self, sample_wrapper, **config):
        """Constructor method"""

        # Set the sample and the extractor configuration
        self.wrapper = sample_wrapper
        self.config = config if config else {}

        # Set the features to be skipped during preparation
        self.skip_features = self.config.pop("skip_features", ("statistics", ))

        # Set the logging
        self.logging_settings = self.config.pop("logging_settings", {})

    # ------------------------------- #
    # Alternative constructor methods #
    # ------------------------------- #

    @classmethod
    def from_sample(cls, sample, **config):
        """
        Initializes HandwritingFeatures object from a sample.
        :param sample: handwriting sample object
        :type sample: HandwritingSample
        :param config: common configuration
        :type config: **kwargs
        :return: HandwritingFeatures object
        :rtype: HandwritingFeatures
        """
        return cls(HandwritingSampleWrapper(sample), **config)

    @classmethod
    def from_list(cls, values, labels=None, **config):
        """
        Initializes HandwritingFeatures object from a list.

        :param values: data values
        :type values: list
        :param labels: labels for the data values
        :type labels: list, optional
        :param config: common configuration
        :type config: **kwargs
        :return: HandwritingFeatures object
        :rtype: HandwritingFeatures
        """
        return cls(HandwritingSampleWrapper.from_list(values, labels), **config)

    @classmethod
    def from_numpy_array(cls, values, labels=None, **config):
        """
        Initializes HandwritingFeatures object from a numpy array.

        :param values: data values
        :type values: numpy.ndarray
        :param labels: labels for the data values
        :type labels: list, optional
        :param config: common configuration
        :type config: **kwargs
        :return: HandwritingFeatures object
        :rtype: HandwritingFeatures
        """
        return cls(HandwritingSampleWrapper.from_numpy_array(values, labels), **config)

    @classmethod
    def from_pandas_dataframe(cls, values, labels=None, **config):
        """
        Initializes HandwritingFeatures object from a pandas DataFrame.

        :param values: data values
        :type values: pandas.DataFrame
        :param labels: labels for the data values
        :type labels: list, optional
        :param config: common configuration
        :type config: **kwargs
        :return: HandwritingFeatures object
        :rtype: HandwritingFeatures
        """
        return cls(HandwritingSampleWrapper.from_pandas_dataframe(values, labels), **config)

    @classmethod
    def from_json(cls, path, labels=None, **config):
        """
        Initializes HandwritingFeatures object from a JSON file.

        :param path: path to a JSON file
        :type path: str
        :param labels: labels for the data values
        :type labels: list, optional
        :param config: common configuration
        :type config: **kwargs
        :return: HandwritingFeatures object
        :rtype: HandwritingFeatures
        """
        return cls(HandwritingSampleWrapper.from_json(path, labels), **config)

    @classmethod
    def from_svc(cls, path, labels=None, **config):
        """
        Initializes HandwritingFeatures object from an SVC file.

        :param path: path to an SVC file
        :type path: str
        :param labels: labels for the data values
        :type labels: list, optional
        :param config: common configuration
        :type config: **kwargs
        :return: HandwritingFeatures object
        :rtype: HandwritingFeatures
        """
        return cls(HandwritingSampleWrapper.from_svc(path, labels), **config)

    # ----------------- #
    # Computation hooks #
    # ----------------- #

    @classmethod
    def inject_common_configuration(cls, method, kwargs, common, skip_features=None):
        """
        Injects the common configuration into the kwargs.

        :param method: feature computation method to be applied
        :type method: callable
        :param kwargs: feature computation-specific kwargs
        :type kwargs: dict
        :param skip_features: features to be skipped during injection, defaults to None
        :type skip_features: Any[list, tuple], optional
        :param common: common configuration
        :type common: dict
        :return: kwargs
        :rtype: dict
        """
        return HandwritingFeaturesFusion.fuze(method.__name__, kwargs, common, skip_features)

    @classmethod
    def before_computation(cls, method, kwargs=None, skip_features=None):
        """
        Applies the before-computation hook.

        :param method: feature computation method to be applied
        :type method: callable
        :param kwargs: feature computation-specific kwargs
        :type kwargs: dict
        :param skip_features: features to be skipped during validation, defaults to None
        :type skip_features: Any[list, tuple], optional
        :return: kwargs
        :rtype: dict
        """
        return HandwritingFeaturesValidation.validate(method.__name__, kwargs, skip_features)

    @classmethod
    def after_computation(cls, feature, statistics=None):
        """
        Applies the after-computation hook.

        :param feature: computed feature
        :type feature: numpy.ndarray
        :param statistics: statistics to be computed, defaults to None
        :type statistics: Any[str, list, tuple], optional
        :return: features
        :rtype: numpy.ndarray
        """

        # Handle the statistics options
        statistics = [statistics] if isinstance(statistics, str) else statistics

        # Compute the statistics
        if statistics:
            feature = numpy.array([Statistics.compute(feature, stat) for stat in statistics])
        else:
            if not isinstance(feature, numpy.ndarray):
                feature = numpy.array(feature).reshape((1,))

        # Return the feature
        return feature

    def compute(self, sample_wrapper, method, statistics=None, **kwargs):
        """
        Applies the computation (including before-after computational hooks).

        :param sample_wrapper: sample wrapper object
        :type sample_wrapper: HandwritingSampleWrapper
        :param method: feature computation method to be applied
        :type method: callable
        :param statistics: statistics to be computed, defaults to None
        :type statistics: iterable, optional
        :param kwargs: feature computation-specific kwargs
        :type kwargs: **kwargs
        :return: computed features
        :rtype: numpy.ndarray
        """

        # Prepare the method name
        method_name = method.__name__

        # Prepare the configuration
        settings = self.inject_common_configuration(method, kwargs, self.config, self.skip_features)

        # Apply the before-computation hook
        settings = self.before_computation(method, settings, self.skip_features)

        # Apply the computation
        try:
            features = method(sample_wrapper, **settings)
        except Exception as e:
            if self.logging_settings.get("soft_validation"):
                logger.warning(f"Failure '{method_name}' for {sample_wrapper}: {e.__class__.__name__} - {e}")
                if HandwritingFeaturesSettings.is_feature_multivalued(method_name):
                    features = numpy.array([])
                else:
                    features = numpy.nan
            else:
                raise e

        # Apply the after-computation hook
        features = self.after_computation(features, statistics)

        # Return the features
        return features



import numpy
import warnings
from handwriting_features.data.utils.cleanup import remove_outliers
from handwriting_features.data.exceptions.statistics import *


# Ignore numpy warnings
warnings.filterwarnings("ignore")


def mean(array):
    """
    Computes mean of an input array (ignoring NaNs).

    :param array: input array
    :type array: numpy.ndarray
    :return: mean value
    :rtype: numpy.float
    """
    return numpy.nanmean(array) if numpy.isfinite(array).any() else numpy.nan


def std(array):
    """
    Computes std of an input array (ignoring NaNs).

    :param array: input array
    :type array: numpy.ndarray
    :return: std value
    :rtype: numpy.float
    """
    return numpy.nanstd(array) if numpy.isfinite(array).any() else numpy.nan


def cv_parametric(array, as_percentage=False):
    """
    Computes parametric cv of an input array (ignoring NaNs).

    :param array: input array
    :type array: numpy.ndarray
    :param as_percentage: percentage computation flag, defaults to False
    :type as_percentage: bool, optional
    :return: parametric cv value
    :rtype: numpy.float
    """

    # Get the mean and standard deviation
    _avg = mean(array)
    _std = std(array)

    # Compute the parametric cv
    if any((not numpy.isfinite(_avg), not numpy.isfinite(_std))):
        return numpy.nan
    else:
        return (_std / (_avg + numpy.finfo(float).eps)) * (1 if not as_percentage else 100)


def median(array):
    """
    Computes median of an input array (ignoring NaNs).

    :param array: input array
    :type array: numpy.ndarray
    :return: median value
    :rtype: numpy.float
    """
    return numpy.nanmedian(array) if numpy.isfinite(array).any() else numpy.nan


def iqr(array):
    """
    Computes iqr of an input array (ignoring NaNs).

    :param array: input array
    :type array: numpy.ndarray
    :return: iqr value
    :rtype: numpy.float
    """

    # Get the quartiles
    _q1 = numpy.nanquantile(array, 0.25)
    _q3 = numpy.nanquantile(array, 0.75)

    # Compute the iqr
    if any((not numpy.isfinite(_q1), not numpy.isfinite(_q3))):
        return numpy.nan
    else:
        return numpy.subtract(_q3, _q1)


def cv_nonparametric(array, as_percentage=False):
    """
    Computes non-parametric cv of an input array (ignoring NaNs).

    :param array: input array
    :type array: numpy.ndarray
    :param as_percentage: percentage computation flag, defaults to False
    :type as_percentage: bool, optional
    :return: non-parametric cv value
    :rtype: numpy.float
    """

    # Get the iqr and median
    _med = median(array)
    _iqr = iqr(array)

    # Compute the non-parametric cv
    if any((not numpy.isfinite(_med), not numpy.isfinite(_iqr))):
        return numpy.nan
    else:
        return (_iqr / _med + numpy.finfo(float).eps) * (1 if not as_percentage else 100)


def quartile_1(array):
    """
    Computes 1st quartile of an input array (ignoring NaNs).

    :param array: input array
    :type array: numpy.ndarray
    :return: 1st quartile value
    :rtype: numpy.float
    """
    return numpy.nanquantile(array, 0.25) if numpy.isfinite(array).any() else numpy.nan


def quartile_3(array):
    """
    Computes 3rd quartile of an input array (ignoring NaNs).

    :param array: input array
    :type array: numpy.ndarray
    :return: 3rd quartile value
    :rtype: numpy.float
    """
    return numpy.nanquantile(array, 0.75) if numpy.isfinite(array).any() else numpy.nan


def percentile_5(array):
    """
    Computes 5th percentile of an input array (ignoring NaNs).

    :param array: input array
    :type array: numpy.ndarray
    :return: 5th percentile value
    :rtype: numpy.float
    """
    return numpy.nanpercentile(array, 5) if numpy.isfinite(array).any() else numpy.nan


def percentile_95(array):
    """
    Computes 95th percentile of an input array (ignoring NaNs).

    :param array: input array
    :type array: numpy.ndarray
    :return: 95th percentile value
    :rtype: numpy.float
    """
    return numpy.nanpercentile(array, 95) if numpy.isfinite(array).any() else numpy.nan


def slope_of_linear_regression(array, window_size=5, min_samples=3, center=True, threshold=3):
    """
    Computes slope of linear regression of an input array (ignoring NaNs).

    :param array: input array
    :type array: numpy.ndarray
    :param window_size: size of the moving window, defaults to 5
    :type window_size: int, optional
    :param min_samples: minimum number of samples in a window, defaults to 3
    :type min_samples: int, optional
    :param center: labels at the center of the window flag, defaults to True
    :type center: bool, optional
    :param threshold: outlier removal threshold, defaults to 3
    :type threshold: int, optional
    :return: slope of linear regression value
    :rtype: numpy.float
    """

    # Handle NaN/Inf value only
    if not isinstance(array, numpy.ndarray) and not numpy.isfinite(array):
        return numpy.nan

    # Create a local deepcopy of the input array
    data = numpy.copy(array)

    # Remove NaN/Inf values and flatten the array
    data = data[numpy.isfinite(data)]
    data = data.flatten("F")

    # Remove outliers
    if data.size > window_size:
        data = remove_outliers(data, window_size, min_samples=min_samples, center=center, threshold=threshold)

    # Handle empty array
    if data.size == 0:
        return numpy.nan

    # Fit the 1st order regression curve
    try:
        fitted = numpy.polyfit(list(range(len(data))), data, 1)
    except numpy.linalg.LinAlgError:
        return numpy.nan

    # Return the slope
    return fitted[0]


class Statistics(object):
    """Class implementing statistics computation interface"""

    # Mapping between statistics and computational functions
    mapping = {
        "mean": mean,
        "std": std,
        "cv_parametric": cv_parametric,
        "median": median,
        "iqr": iqr,
        "cv_nonparametric": cv_nonparametric,
        "quartile_1": quartile_1,
        "quartile_3": quartile_3,
        "percentile_5": percentile_5,
        "percentile_95": percentile_95,
        "slope_of_linear_regression": slope_of_linear_regression
    }

    @classmethod
    def compute(cls, array, statistical_function):
        """
        Computes the <statistical_function> of an input <array>.

        :param array: input array
        :type array: numpy.ndarray
        :param statistical_function: statistical function name
        :type statistical_function: str
        :return: computed statistics
        :rtype: numpy.float
        """

        # Validate input arguments
        if statistical_function not in cls.mapping:
            raise StatisticsNameNotInMappingError(f"Unsupported <statistical_function> {statistical_function}")
        if not isinstance(array, (float, numpy.ndarray)):
            raise UnsupportedDataForStatisticsError(
                f"Unsupported <array> type {type(array)}; "
                f"must be any of the following: `numpy.ndarray`, `numpy.float`")

        # Compute the statistical function
        return cls.mapping[statistical_function](array)



import numpy


def stroke_duration(sample_wrapper, in_air):
    """
    Returns stroke duration.

    :param sample_wrapper: sample wrapper object
    :type sample_wrapper: HandwritingSampleWrapper
    :param in_air: in-air flag
    :type in_air: bool
    :return: stroke duration
    :rtype: numpy.ndarray or np.NaN
    """

    # Get the strokes
    strokes = sample_wrapper.on_surface_strokes \
        if not in_air \
        else sample_wrapper.in_air_strokes

    # Check the presence of strokes
    if not strokes:
        return numpy.nan

    # Return the stokes duration
    return numpy.array([max(stroke.time) - min(stroke.time) for stroke in strokes])


def ratio_of_stroke_durations(sample_wrapper):
    """
    Returns ratio of stroke durations: on-surface / in-air.

    :param sample_wrapper: sample wrapper object
    :type sample_wrapper: HandwritingSampleWrapper
    :return: ratio of stroke durations
    :rtype: numpy.ndarray or np.NaN
    """

    # Get on-surface and in-air stroke durations
    on_surface_strokes = stroke_duration(sample_wrapper, in_air=False)
    in_air_strokes = stroke_duration(sample_wrapper, in_air=True)

    # Check the presence of the stroke durations
    if any((not isinstance(on_surface_strokes, numpy.ndarray), not isinstance(in_air_strokes, numpy.ndarray))):
        return numpy.nan

    # Get the ratio between the stroke durations
    ratio = [(x / (y + numpy.finfo(float).eps)) for x, y in zip(on_surface_strokes, in_air_strokes)]
    ratio = numpy.array(ratio) if ratio else numpy.nan

    # Return the ratio
    return ratio


def writing_duration(sample_wrapper, in_air):
    """
    Returns writing duration.

    :param sample_wrapper: sample wrapper object
    :type sample_wrapper: HandwritingSampleWrapper
    :param in_air: in-air flag
    :type in_air: bool
    :return: writing duration
    :rtype: float
    """
    return numpy.sum(stroke_duration(sample_wrapper, in_air))


def writing_duration_overall(sample_wrapper):
    """
    Returns overall writing duration.

    :param sample_wrapper: sample wrapper object
    :type sample_wrapper: HandwritingSampleWrapper
    :return: overall writing duration
    :rtype: float
    """
    return float(sample_wrapper.sample_time[-1] - sample_wrapper.sample_time[0])


def ratio_of_writing_durations(sample_wrapper):
    """
    Returns ratio of writing durations: on-surface / in-air.

    :param sample_wrapper: sample wrapper object
    :type sample_wrapper: HandwritingSampleWrapper
    :return: ratio of writing durations
    :rtype: float
    """

    # Get on-surface and in-air writing durations
    on_surface_writing_duration = writing_duration(sample_wrapper, in_air=False)
    in_air_writing_duration = writing_duration(sample_wrapper, in_air=True)

    # Check the presence of the writing durations
    if any((not numpy.isfinite(on_surface_writing_duration), not numpy.isfinite(in_air_writing_duration))):
        return numpy.nan

    # Return the ratio between the writing durations
    return on_surface_writing_duration / (in_air_writing_duration + numpy.finfo(float).eps)


def number_of_interruptions(sample_wrapper):
    """
    Returns the number of interruptions.

    :param sample_wrapper: sample wrapper object
    :type sample_wrapper: HandwritingSampleWrapper
    :return: number of interruptions
    :rtype: float
    """

    # Get the pen status
    pen_status = sample_wrapper.sample_pen_status

    # Return the number of interruptions
    return float(max(sum(abs(numpy.logical_xor(pen_status[:-1], pen_status[1:]))), 0))


def number_of_interruptions_relative(sample_wrapper):
    """
    Returns the number of interruptions relative to the duration.

    :param sample_wrapper: sample wrapper object
    :type sample_wrapper: HandwritingSampleWrapper
    :return: number of interruptions relative to the duration
    :rtype: float
    """

    # Get the number of interruptions and the overal writing duration
    interruptions = number_of_interruptions(sample_wrapper)
    duration = writing_duration_overall(sample_wrapper)

    # Return the number of interruptions relative to the duration
    return interruptions / (duration + numpy.finfo(float).eps)



import numpy
from itertools import chain, product
from handwriting_features.features.configuration.settings import HandwritingFeaturesSettings


class SingleSubjectFeatureUtils(object):
    """Class implementing single-subject feature values/labels utils"""

    # Handwriting features settings
    features_settings = HandwritingFeaturesSettings

    @classmethod
    def prepare_feature_values(cls, extracted):
        """
        Prepares the feature values.

        :param extracted: extracted feature values
        :type extracted: numpy.ndarray
        :return: prepared feature values
        :rtype: numpy.ndarray
        """
        return extracted

    @classmethod
    def prepare_feature_labels(cls, extracted, feature_name, feature_args=None):
        """
        Prepares the feature labels.

        :param extracted: extracted feature values
        :type extracted: numpy.ndarray
        :param feature_name: feature name
        :type feature_name: str
        :param feature_args: feature arguments, defaults to None
        :type feature_args: dict, optional
        :return: prepared feature labels
        :rtype: list
        """

        # Prepare the base of the feature label(s) (handle these cases)
        #
        # 1. feature with statistics
        # 2. feature without statistics
        #    a) single-valued feature (i.e. no statistics are available)
        #    b) multi-valued (array-like) feature

        # Prepare default feature args
        feature_args = feature_args if feature_args else {}

        # 1. Handle: feature with statistics
        if feature_args.get("statistics"):

            # Prepare the feature label(s)
            statistics = feature_args["statistics"]
            statistics = [statistics] if isinstance(statistics, str) else statistics

            labels = [
                f"{stat}:{feature_name}"
                for stat in statistics
            ]

        # 2. Handle: feature without statistics
        else:

            # Prepare the feature label(s)
            labels = [f"{feature_name}"] * len(extracted)

            # Handle multi-values feature
            if len(labels) > 1:
                preset = None
                if cls.features_settings.settings.get(feature_name, {}).get("properties", {}).get("is_composite"):
                    preset = cls.features_settings.settings[feature_name]["properties"].get("composite_feature_names")
                    if preset:
                        labels = preset \
                            if len(preset) == len(labels) \
                            else preset + [f"{label}(missing name specification)" for label in labels[len(preset):]]
                if not preset:
                    labels = [f"{label}(sample-{i})" for i, label in enumerate(labels, 1)]

        # Add the axis information
        if feature_args.get("axis"):
            if cls.features_settings.settings.get(feature_name, {}).get("arguments").get("axis"):
                labels = [
                    f"{label}:axis-{feature_args['axis']}"
                    for label in labels
                ]
        else:
            option = cls.features_settings.get_feature_argument_default(feature_name, "axis")
            labels = [f"{label}:axis-{option}" for label in labels] \
                if option is not None \
                else labels

        # Add the in-air/on-surface information
        if feature_args.get("in_air") is not None:
            if cls.features_settings.settings.get(feature_name, {}).get("arguments").get("in_air"):
                labels = [
                    f"{label}({'in-air' if feature_args['in_air'] else 'on-surface'})"
                    for label in labels
                ]
        else:
            option = cls.features_settings.get_feature_argument_default(feature_name, "in_air")
            labels = [f"{label}({'in-air' if option else 'on-surface'})" for label in labels] \
                if option is not None \
                else labels

        # Return the prepared feature labels
        return labels


class MultiSubjectFeatureUtils(object):
    """Class implementing multi-subject feature values/labels utils"""

    @classmethod
    def prepare_feature_values(cls, extracted, pipeline):
        """
        Prepares the feature values.

        :param extracted: extracted features
        :type extracted: list
        :param pipeline: pipeline of the features to be extracted
        :type pipeline: list
        :return: finalized feature values
        :rtype: numpy.ndarray
        """

        # Get the number of subjects and features
        num_subjects = len(extracted)
        num_features = len(pipeline)

        # Prepare the list of feature values (each feature with data for all subjects)
        features = [
            [subject["features"][feature] for subject in extracted]
            for feature in range(num_features)
        ]

        # Prepare the list of consolidated feature values
        consolidated = []

        # Consolidate the lengths of the feature values
        for feature in features:
            max_length = max(len(subject) for subject in feature)

            if all(len(subject) == max_length for subject in feature):
                consolidated.append(feature)
            else:
                consolidated.append([
                    numpy.concatenate([subject, numpy.full((max_length - len(subject),), numpy.nan)])
                    for subject in feature
                ])

        # Prepare the feature values to be finalized
        features = [
            [feature[subject] for feature in consolidated]
            for subject in range(num_subjects)
        ]

        # Finalize the feature values (concatenate: 1. over features. 2. over subjects)
        features = [numpy.expand_dims(numpy.hstack(subject), axis=0) for subject in features]
        features = numpy.vstack(list(subject for subject in features))

        # Return the finalized feature values
        return features

    @classmethod
    def prepare_feature_labels(cls, extracted, pipeline):
        """
        Prepares the feature labels.

        :param extracted: extracted features
        :type extracted: list
        :param pipeline: pipeline of the features to be extracted
        :type pipeline: list
        :return: finalized feature labels
        :rtype: list
        """

        # Get the number of features
        num_features = len(pipeline)

        # Prepare the list of feature labels (each feature with data for all subjects)
        features = [
            [subject["labels"][feature] for subject in extracted]
            for feature in range(num_features)
        ]

        # Finalize the feature labels
        labels = list(chain.from_iterable([max(feature, key=len) for feature in features]))

        # Return the finalized feature labels
        return labels


class FeaturesPipelineUtils(object):
    """Class implementing features pipeline utils"""

    # Arguments to be prepared
    arguments_to_prepare = ["axis", "in_air"]

    @classmethod
    def prepare_features_pipeline(cls, pipeline):
        """
        Prepares the features pipeline.

        :param pipeline: pipeline of the features to be extracted
        :type pipeline: list
        :return: prepared pipeline
        :rtype: list
        """

        # Prepare an empty updated features pipeline
        prepared = []

        # Prepare the features pipeline
        for feature in pipeline:
            name, args = feature.get("name"), feature.get("args", {})

            # Clean the arguments
            args = {key: val for key, val in args.items() if val}

            # Separate the arguments not/to be prepared
            unchanged_args = {key: val for key, val in args.items() if key not in cls.arguments_to_prepare}
            processed_args = [
                (key, val if isinstance(val, list) else [val])
                for key, val in args.items() if key in cls.arguments_to_prepare
            ]

            # Process the arguments
            processed_args = list(product(*[[(key, v) for v in val] for key, val in processed_args]))

            # Add the feature to the features pipeline
            for changed_args in processed_args:
                prepared.append({
                    "name": name,
                    "args": {**unchanged_args, **dict(changed_args)}
                })

        # Return the prepared features pipeline
        return prepared



from itertools import islice
from collections import deque


def sliding_window(iterable, size, step, padding=None):
    """
    Apply sliding_window on an input iterable given the windowing arguments.

    :param iterable: iterable to be iterated over
    :type iterable: iterable
    :param size: window size
    :type size: int
    :param step: window step
    :type step: int
    :param padding: window padding, defaults to None
    :type padding: Any, optional
    :return: windowed iterable
    :rtype: iterator
    """

    # Prepare the queue
    i = iter(iterable)
    q = deque(islice(i, size), maxlen=size)
    if not q:
        return

    # Apply the window
    q.extend(padding for _ in range(size - len(q)))
    while True:
        yield iter(q)
        try:
            q.append(next(i))
        except StopIteration:
            return
        q.extend(next(i, padding) for _ in range(step - 1))



import numpy
import functools
from handwriting_sample import HandwritingSample
from handwriting_features.data.utils.math import derivation
from handwriting_features.data.exceptions.sample import *


class HandwritingSampleWrapper(object):
    """Class implementing the handwriting sample wrapper"""

    # Handwriting data axes
    axes = ("x", "y", "xy")

    # Handwriting data surface information
    surfaces = ("on_surface", "in_air")

    def __init__(self, sample, source=None):
        """Constructor method"""

        # Set the sample
        self.sample = sample
        self.source = source

        # Set the strokes
        self.strokes = self.sample.get_strokes()

    def __str__(self):
        return f"{self.source}" if self.source else f"HandwritingSampleWrapper({self.sample})"

    def __repr__(self):
        return self.__str__()

    # ------------------------------- #
    # Alternative constructor methods #
    # ------------------------------- #

    @classmethod
    def from_list(cls, values, labels=None):
        """
        Initializes HandwritingSampleWrapper object from a list.

        :param values: data values
        :type values: list
        :param labels: labels for the data values
        :type labels: list, optional
        :return: HandwritingSampleWrapper object
        :rtype: HandwritingSampleWrapper
        """
        return cls(HandwritingSample.from_list(values, labels))

    @classmethod
    def from_numpy_array(cls, values, labels=None):
        """
        Initializes HandwritingSampleWrapper object from a numpy array.

        :param values: data values
        :type values: numpy.ndarray
        :param labels: labels for the data values
        :type labels: list, optional
        :return: HandwritingSampleWrapper object
        :rtype: HandwritingSampleWrapper
        """
        return cls(HandwritingSample.from_numpy_array(values, labels))

    @classmethod
    def from_pandas_dataframe(cls, values, labels=None):
        """
        Initializes HandwritingSampleWrapper object from a pandas DataFrame.

        :param values: data values
        :type values: pandas.DataFrame
        :param labels: labels for the data values
        :type labels: list, optional
        :return: HandwritingSampleWrapper object
        :rtype: HandwritingSampleWrapper
        """
        return cls(HandwritingSample.from_pandas_dataframe(values, labels))

    @classmethod
    def from_json(cls, path, labels=None):
        """
        Initializes HandwritingSampleWrapper object from a JSON file.

        :param path: path to a JSON file
        :type path: str
        :param labels: labels for the data values
        :type labels: list, optional
        :return: HandwritingSampleWrapper object
        :rtype: HandwritingSampleWrapper
        """
        return cls(HandwritingSample.from_json(path, labels), path)

    @classmethod
    def from_svc(cls, path, labels=None):
        """
        Initializes HandwritingSampleWrapper object from an SVC file.

        :param path: path to an SVC file
        :type path: str
        :param labels: labels for the data values
        :type labels: list, optional
        :return: HandwritingSampleWrapper object
        :rtype: HandwritingSampleWrapper
        """
        return cls(HandwritingSample.from_svc(path, labels), path)

    # ----------------------------- #
    # Derived handwriting variables #
    # ----------------------------- #

    @functools.lru_cache(maxsize=len(axes) * len(surfaces))
    def compute_velocity(self, axis="xy", in_air=False):
        """
        Computes the velocity.

        :param axis: axis to compute the velocity from, defaults to "xy"
        :type axis: str, optional
        :param in_air: in-air flag, defaults to False
        :type in_air: bool, optional
        :return: velocity
        :rtype: numpy.ndarray or numpy.NaN
        """

        # Validate the input arguments
        self.validate_axis(axis)
        self.validate_surface_movement(in_air)

        # Compute the velocity
        velocity = numpy.concatenate(self._compute_strokes_velocities(axis, in_air))

        # Return the velocity
        return velocity if velocity is not None and velocity.size > 0 else numpy.nan

    @functools.lru_cache(maxsize=len(axes) * len(surfaces))
    def compute_acceleration(self, axis="xy", in_air=False):
        """
        Computes the acceleration.

        :param axis: axis to compute the acceleration from, defaults to "xy"
        :type axis: str, optional
        :param in_air: in-air flag, defaults to False
        :type in_air: bool, optional
        :return: acceleration
        :rtype: numpy.ndarray or numpy.NaN
        """

        # Validate the input arguments
        self.validate_axis(axis)
        self.validate_surface_movement(in_air)

        # Compute the acceleration
        acceleration = numpy.concatenate(self._compute_strokes_accelerations(axis, in_air))

        # Return the acceleration
        return acceleration if acceleration is not None and acceleration.size > 0 else numpy.nan

    @functools.lru_cache(maxsize=len(axes) * len(surfaces))
    def compute_jerk(self, axis="xy", in_air=False):
        """
        Computes the jerk.

        :param axis: axis to compute the jerk from, defaults to "xy"
        :type axis: str, optional
        :param in_air: in-air flag, defaults to False
        :type in_air: bool, optional
        :return: jerk
        :rtype: numpy.ndarray or numpy.NaN
        """

        # Validate the input arguments
        self.validate_axis(axis)
        self.validate_surface_movement(in_air)

        # Compute the jerk
        jerk = numpy.concatenate(self._compute_strokes_jerks(axis, in_air))

        # Return the jerk
        return jerk if jerk is not None and jerk.size > 0 else numpy.nan

    @functools.lru_cache(maxsize=len(surfaces))
    def compute_azimuth(self, in_air=False):
        """
        Computes the azimuth.

        :param in_air: in-air flag, defaults to False
        :type in_air: bool, optional
        :return: azimuth
        :rtype: numpy.ndarray or numpy.NaN
        """

        # Validate the input arguments
        self.validate_surface_movement(in_air)

        # Get the data to be used for the feature computation
        data = self.on_surface_data if not in_air else self.in_air_data

        # Return the azimuth
        return data.azimuth

    @functools.lru_cache(maxsize=len(surfaces))
    def compute_tilt(self, in_air=False):
        """
        Computes the tilt.

        :param in_air: in-air flag, defaults to False
        :type in_air: bool, optional
        :return: tilt
        :rtype: numpy.ndarray or numpy.NaN
        """

        # Validate the input arguments
        self.validate_surface_movement(in_air)

        # Get the data to be used for the feature computation
        data = self.on_surface_data if not in_air else self.in_air_data

        # Return the tilt
        return data.tilt

    @functools.lru_cache(maxsize=1)
    def compute_pressure(self):
        """
        Computes the pressure.

        :return: tilt
        :rtype: numpy.ndarray or numpy.NaN
        """
        return self.on_surface_data.pressure

    # ---------------------------- #
    # Sample handwriting variables #
    # ---------------------------- #

    @property
    def sample_x(self):
        return self.sample.x

    @property
    def sample_y(self):
        return self.sample.y

    @property
    def sample_time(self):
        return self.sample.time

    @property
    def sample_pen_status(self):
        return self.sample.pen_status

    @property
    def sample_azimuth(self):
        return self.sample.azimuth

    @property
    def sample_tilt(self):
        return self.sample.tilt

    @property
    def sample_pressure(self):
        return self.sample.pressure

    @functools.cached_property
    def in_air_data(self):
        return self.sample.get_in_air_data()

    @functools.cached_property
    def on_surface_data(self):
        return self.sample.get_on_surface_data()

    @functools.cached_property
    def in_air_strokes(self):
        return [stroke for status, stroke in self.strokes if status == "in_air"]

    @functools.cached_property
    def on_surface_strokes(self):
        return [stroke for status, stroke in self.strokes if status == "on_surface"]

    # ------------------- #
    # Validation routines #
    # ------------------- #

    @classmethod
    def validate_axis(cls, axis):
        """Validates the axis"""
        if axis not in cls.axes:
            raise UnsupportedAxisError(f"Unsupported <axis> argument {axis}; must be in {cls.axes}")

    @classmethod
    def validate_surface_movement(cls, in_air):
        """Validates the surface movement"""
        if not isinstance(in_air, bool):
            raise UnsupportedSurfaceMovementError(f"Unsupported <in_air> argument {in_air}; must be bool")

    # ---------------------- #
    # Computational routines #
    # ---------------------- #

    @functools.lru_cache(maxsize=len(axes) * len(surfaces))
    def _compute_strokes_trajectories(self, axis="xy", in_air=False):
        """
        Computes the strokes trajectories.

        :param axis: axis to compute the trajectories from, defaults to "xy"
        :type axis: str, optional
        :param in_air: in-air flag, defaults to False
        :type in_air: bool, optional
        :return: strokes trajectories
        :rtype: list
        """

        # Get the strokes
        strokes = self.on_surface_strokes if not in_air else self.in_air_strokes

        # Return the strokes trajectories
        if axis == "x":
            return [numpy.abs(derivation(stroke.x)) for stroke in strokes]
        if axis == "y":
            return [numpy.abs(derivation(stroke.y)) for stroke in strokes]
        if axis == "xy":
            return [
                numpy.sqrt(numpy.power(derivation(stroke.x), 2) + numpy.power(derivation(stroke.y), 2))
                for stroke in strokes
            ]

    @functools.lru_cache(maxsize=len(surfaces))
    def _compute_strokes_time_differences(self, in_air=False):
        """
        Computes the strokes time differences.

        :param in_air: in-air flag, defaults to False
        :type in_air: bool, optional
        :return: strokes time differences
        :rtype: list
        """
        return [
            derivation(stroke.time)
            for stroke in (self.on_surface_strokes if not in_air else self.in_air_strokes)
        ]

    @functools.lru_cache(maxsize=len(axes) * len(surfaces))
    def _compute_strokes_velocities(self, axis="xy", in_air=False):
        """
        Computes the strokes velocities.

        :param axis: axis to compute the velocities from, defaults to "xy"
        :type axis: str, optional
        :param in_air: in-air flag, defaults to False
        :type in_air: bool, optional
        :return: strokes velocities
        :rtype: list
        """

        # Get the variable to differentiate over (strokes trajectories)
        ds = self._compute_strokes_trajectories(axis, in_air)

        # Get the stroke time differences
        dt = self._compute_strokes_time_differences(in_air)

        # Return the stokes velocities
        return [d / t for (d, t) in zip(ds, dt)]

    @functools.lru_cache(maxsize=len(axes) * len(surfaces))
    def _compute_strokes_accelerations(self, axis="xy", in_air=False):
        """
        Computes the strokes accelerations.

        :param axis: axis to compute the accelerations from, defaults to "xy"
        :type axis: str, optional
        :param in_air: in-air flag, defaults to False
        :type in_air: bool, optional
        :return: strokes accelerations
        :rtype: list
        """

        # Get the variable to differentiate over (strokes velocities)
        dv = self._compute_strokes_velocities(axis, in_air)

        # Get the stroke time differences
        dt = self._compute_strokes_time_differences(in_air)

        # Return the stokes accelerations
        return [derivation(d) / t[1:] for (d, t) in zip(dv, dt)]

    @functools.lru_cache(maxsize=len(axes) * len(surfaces))
    def _compute_strokes_jerks(self, axis="xy", in_air=False):
        """
        Computes the strokes jerks.

        :param axis: axis to compute the jerks from, defaults to "xy"
        :type axis: str, optional
        :param in_air: in-air flag, defaults to False
        :type in_air: bool, optional
        :return: strokes jerks
        :rtype: list
        """

        # Get the variable to differentiate over (strokes accelerations)
        da = self._compute_strokes_accelerations(axis, in_air)

        # Get the stroke time differences
        dt = self._compute_strokes_time_differences(in_air)

        # Return the stokes jerks
        return [derivation(d) / t[2:] for (d, t) in zip(da, dt)]



from handwriting_features.features.base import HandwritingFeaturesBase
from handwriting_features.features.implementation import *


class HandwritingFeatures(HandwritingFeaturesBase):
    """Class implementing interface for computation of handwriting features"""

    # -------------------- #
    # Handwriting features #
    # -------------------- #

    # ---------------------
    # 1. Kinematic features

    def velocity(self, axis="xy", in_air=False, statistics=()):
        """
        Extracts the velocity.

        :param axis: axis to compute the velocity from, defaults to "xy"
        :type axis: str, optional
        :param in_air: in-air flag, defaults to False
        :type in_air: bool, optional
        :param statistics: statistics to compute, defaults to ()
        :type statistics: Any[list, tuple], optional
        :return: velocity
        :rtype: numpy.ndarray or numpy.NaN
        """
        return self.compute(self.wrapper, velocity, statistics=statistics, axis=axis, in_air=in_air)

    def acceleration(self, axis="xy", in_air=False, statistics=()):
        """
        Extracts the acceleration.

        :param axis: axis to compute the acceleration from, defaults to "xy"
        :type axis: str, optional
        :param in_air: in-air flag, defaults to False
        :type in_air: bool, optional
        :param statistics: statistics to compute, defaults to ()
        :type statistics: Any[list, tuple], optional
        :return: acceleration
        :rtype: numpy.ndarray or numpy.NaN
        """
        return self.compute(self.wrapper, acceleration, statistics=statistics, axis=axis, in_air=in_air)

    def jerk(self, axis="xy", in_air=False, statistics=()):
        """
        Extracts the jerk.

        :param axis: axis to compute the jerk from, defaults to "xy"
        :type axis: str, optional
        :param in_air: in-air flag, defaults to False
        :type in_air: bool, optional
        :param statistics: statistics to compute, defaults to ()
        :type statistics: Any[list, tuple], optional
        :return: jerk
        :rtype: numpy.ndarray or numpy.NaN
        """
        return self.compute(self.wrapper, jerk, statistics=statistics, axis=axis, in_air=in_air)

    # -------------------
    # 2. Dynamic features

    def azimuth(self, in_air=False, statistics=()):
        """
        Extracts the azimuth.

        :param in_air: in-air flag, defaults to False
        :type in_air: bool, optional
        :param statistics: statistics to compute, defaults to ()
        :type statistics: Any[list, tuple], optional
        :return: azimuth
        :rtype: numpy.ndarray or numpy.NaN
        """
        return self.compute(self.wrapper, azimuth, statistics=statistics, in_air=in_air)

    def tilt(self, in_air=False, statistics=()):
        """
        Extracts the tilt.

        :param in_air: in-air flag, defaults to False
        :type in_air: bool, optional
        :param statistics: statistics to compute, defaults to ()
        :type statistics: Any[list, tuple], optional
        :return: tilt
        :rtype: numpy.ndarray or numpy.NaN
        """
        return self.compute(self.wrapper, tilt, statistics=statistics, in_air=in_air)

    def pressure(self, statistics=()):
        """
        Extracts the pressure.

        :param statistics: statistics to compute, defaults to ()
        :type statistics: Any[list, tuple], optional
        :return: pressure
        :rtype: numpy.ndarray or numpy.NaN
        """
        return self.compute(self.wrapper, pressure, statistics=statistics)

    # -------------------
    # 3. Spatial features

    def stroke_length(self, in_air=False, statistics=()):
        """
        Extracts the stroke length.

        :param in_air: in-air flag, defaults to False
        :type in_air: bool, optional
        :param statistics: statistics to compute, defaults to ()
        :type statistics: Any[list, tuple], optional
        :return: stroke length
        :rtype: numpy.ndarray or numpy.NaN
        """
        return self.compute(self.wrapper, stroke_length, statistics=statistics, in_air=in_air)

    def stroke_height(self, in_air=False, statistics=()):
        """
        Extracts the stroke height.

        :param in_air: in-air flag, defaults to False
        :type in_air: bool, optional
        :param statistics: statistics to compute, defaults to ()
        :type statistics: Any[list, tuple], optional
        :return: stroke height
        :rtype: numpy.ndarray or numpy.NaN
        """
        return self.compute(self.wrapper, stroke_height, statistics=statistics, in_air=in_air)

    def stroke_width(self, in_air=False, statistics=()):
        """
        Extracts the stroke width.

        :param in_air: in-air flag, defaults to False
        :type in_air: bool, optional
        :param statistics: statistics to compute, defaults to ()
        :type statistics: Any[list, tuple], optional
        :return: stroke width
        :rtype: numpy.ndarray or numpy.NaN
        """
        return self.compute(self.wrapper, stroke_width, statistics=statistics, in_air=in_air)

    def writing_length(self, in_air=False):
        """
        Extracts the writing length.

        :param in_air: in-air flag, defaults to False
        :type in_air: bool, optional
        :return: writing length
        :rtype: float
        """
        return self.compute(self.wrapper, writing_length, in_air=in_air)

    def writing_height(self, in_air=False):
        """
        Extracts the writing height.

        :param in_air: in-air flag, defaults to False
        :type in_air: bool, optional
        :return: writing height
        :rtype: float
        """
        return self.compute(self.wrapper, writing_height, in_air=in_air)

    def writing_width(self, in_air=False):
        """
        Extracts the writing width.

        :param in_air: in-air flag, defaults to False
        :type in_air: bool, optional
        :return: writing width
        :rtype: float
        """
        return self.compute(self.wrapper, writing_width, in_air=in_air)

    def number_of_intra_stroke_intersections(self, statistics=()):
        """
        Extracts the number of intra-stroke intersections.

        :param statistics: statistics to compute, defaults to ()
        :type statistics: Any[list, tuple], optional
        :return: number of intra-stroke intersections
        :rtype: numpy.ndarray or np.NaN
        """
        return self.compute(self.wrapper, number_of_intra_stroke_intersections, statistics=statistics)

    def relative_number_of_intra_stroke_intersections(self, statistics=()):
        """
        Extracts the relative number of intra-stroke intersections.

        :param statistics: statistics to compute, defaults to ()
        :type statistics: Any[list, tuple], optional
        :return: relative number of intra-stroke intersections
        :rtype: numpy.ndarray or np.NaN
        """
        return self.compute(self.wrapper, relative_number_of_intra_stroke_intersections, statistics=statistics)

    def total_number_of_intra_stroke_intersections(self):
        """
        Extracts the total number of intra-stroke intersections.

        :return: total number of intra-stroke intersections
        :rtype: int
        """
        return self.compute(self.wrapper, total_number_of_intra_stroke_intersections)

    def relative_total_number_of_intra_stroke_intersections(self):
        """
        Extracts the relative total number of intra-stroke intersections.

        :return: relative total number of intra-stroke intersections
        :rtype: float
        """
        return self.compute(self.wrapper, relative_total_number_of_intra_stroke_intersections)

    def number_of_inter_stroke_intersections(self):
        """
        Extracts the number of inter-stroke intersections.

        :return: number of inter-stroke intersections
        :rtype: int
        """
        return self.compute(self.wrapper, number_of_inter_stroke_intersections)

    def relative_number_of_inter_stroke_intersections(self):
        """
        Extracts the relative number of inter-stroke intersections.

        :return: relative number of inter-stroke intersections
        :rtype: float
        """
        return self.compute(self.wrapper, relative_number_of_inter_stroke_intersections)

    def vertical_peaks_indices(self, fs, n=None, statistics=()):
        """
        Extracts the vertical peaks indices.

        :param fs: sampling frequency
        :type fs: float
        :param n: number of samples of a Gaussian filter, defaults to None
        :type n: int, optional
        :param statistics: statistics to compute, defaults to ()
        :type statistics: Any[list, tuple], optional
        :return: vertical peaks indices
        :rtype: numpy.ndarray or np.NaN
        """
        return self.compute(self.wrapper, vertical_peaks_indices, fs=fs, n=n, statistics=statistics)

    def vertical_valleys_indices(self, fs, n=None, statistics=()):
        """
        Extracts the vertical valleys indices.

        :param fs: sampling frequency
        :type fs: float
        :param n: number of samples of a Gaussian filter, defaults to None
        :type n: int, optional
        :param statistics: statistics to compute, defaults to ()
        :type statistics: Any[list, tuple], optional
        :return: vertical valleys indices
        :rtype: numpy.ndarray or np.NaN
        """
        return self.compute(self.wrapper, vertical_valleys_indices, fs=fs, n=n, statistics=statistics)

    def vertical_peaks_values(self, fs, n=None, statistics=()):
        """
        Extracts the vertical peaks values.

        :param fs: sampling frequency
        :type fs: float
        :param n: number of samples of a Gaussian filter, defaults to None
        :type n: int, optional
        :param statistics: statistics to compute, defaults to ()
        :type statistics: Any[list, tuple], optional
        :return: vertical peaks values
        :rtype: numpy.ndarray or np.NaN
        """
        return self.compute(self.wrapper, vertical_peaks_values, fs=fs, n=n, statistics=statistics)

    def vertical_valleys_values(self, fs, n=None, statistics=()):
        """
        Extracts the vertical valleys values.

        :param fs: sampling frequency
        :type fs: float
        :param n: number of samples of a Gaussian filter, defaults to None
        :type n: int, optional
        :param statistics: statistics to compute, defaults to ()
        :type statistics: Any[list, tuple], optional
        :return: vertical valleys values
        :rtype: numpy.ndarray or np.NaN
        """
        return self.compute(self.wrapper, vertical_valleys_values, fs=fs, n=n, statistics=statistics)

    def vertical_peaks_velocity(self, fs, n=None, statistics=()):
        """
        Extracts the vertical peaks velocity.

        :param fs: sampling frequency
        :type fs: float
        :param n: number of samples of a Gaussian filter, defaults to None
        :type n: int, optional
        :param statistics: statistics to compute, defaults to ()
        :type statistics: Any[list, tuple], optional
        :return: vertical peaks velocity
        :rtype: numpy.ndarray or np.NaN
        """
        return self.compute(self.wrapper, vertical_peaks_velocity, fs=fs, n=n, statistics=statistics)

    def vertical_valleys_velocity(self, fs, n=None, statistics=()):
        """
        Extracts the vertical valleys velocity.

        :param fs: sampling frequency
        :type fs: float
        :param n: number of samples of a Gaussian filter, defaults to None
        :type n: int, optional
        :param statistics: statistics to compute, defaults to ()
        :type statistics: Any[list, tuple], optional
        :return: vertical valleys velocity
        :rtype: numpy.ndarray or np.NaN
        """
        return self.compute(self.wrapper, vertical_valleys_velocity, fs=fs, n=n, statistics=statistics)

    def vertical_peaks_distance(self, fs, n=None, statistics=()):
        """
        Extracts the vertical peaks distance.

        :param fs: sampling frequency
        :type fs: float
        :param n: number of samples of a Gaussian filter, defaults to None
        :type n: int, optional
        :param statistics: statistics to compute, defaults to ()
        :type statistics: Any[list, tuple], optional
        :return: vertical peaks distance
        :rtype: numpy.ndarray or np.NaN
        """
        return self.compute(self.wrapper, vertical_peaks_distance, fs=fs, n=n, statistics=statistics)

    def vertical_valleys_distance(self, fs, n=None, statistics=()):
        """
        Extracts the vertical valleys distance.

        :param fs: sampling frequency
        :type fs: float
        :param n: number of samples of a Gaussian filter, defaults to None
        :type n: int, optional
        :param statistics: statistics to compute, defaults to ()
        :type statistics: Any[list, tuple], optional
        :return: vertical valleys distance
        :rtype: numpy.ndarray or np.NaN
        """
        return self.compute(self.wrapper, vertical_valleys_distance, fs=fs, n=n, statistics=statistics)

    def vertical_peaks_duration(self, fs, n=None, statistics=()):
        """
        Extracts the vertical peaks duration.

        :param fs: sampling frequency
        :type fs: float
        :param n: number of samples of a Gaussian filter, defaults to None
        :type n: int, optional
        :param statistics: statistics to compute, defaults to ()
        :type statistics: Any[list, tuple], optional
        :return: vertical peaks duration
        :rtype: numpy.ndarray or np.NaN
        """
        return self.compute(self.wrapper, vertical_peaks_duration, fs=fs, n=n, statistics=statistics)

    def vertical_valleys_duration(self, fs, n=None, statistics=()):
        """
        Extracts the vertical valleys duration.

        :param fs: sampling frequency
        :type fs: float
        :param n: number of samples of a Gaussian filter, defaults to None
        :type n: int, optional
        :param statistics: statistics to compute, defaults to ()
        :type statistics: Any[list, tuple], optional
        :return: vertical valleys duration
        :rtype: numpy.ndarray or np.NaN
        """
        return self.compute(self.wrapper, vertical_valleys_duration, fs=fs, n=n, statistics=statistics)

    # --------------------
    # 4. Temporal features

    def stroke_duration(self, in_air=False, statistics=()):
        """
        Extracts the stroke duration.

        :param in_air: in-air flag, defaults to False
        :type in_air: bool, optional
        :param statistics: statistics to compute, defaults to ()
        :type statistics: Any[list, tuple], optional
        :return: stroke duration
        :rtype: numpy.ndarray or numpy.NaN
        """
        return self.compute(self.wrapper, stroke_duration, statistics=statistics, in_air=in_air)

    def ratio_of_stroke_durations(self, statistics=()):
        """
        Extracts the ratio of stroke durations: on-surface / in-air.

        :param statistics: statistics to compute, defaults to ()
        :type statistics: Any[list, tuple], optional
        :return: ratio of stroke durations
        :rtype: numpy.ndarray or numpy.NaN
        """
        return self.compute(self.wrapper, ratio_of_stroke_durations, statistics=statistics)

    def writing_duration(self, in_air=False):
        """
        Extracts the writing duration.

        :param in_air: in-air flag, defaults to False
        :type in_air: bool, optional
        :return: writing duration
        :rtype: float
        """
        return self.compute(self.wrapper, writing_duration, in_air=in_air)

    def writing_duration_overall(self):
        """
        Extracts the overall writing duration.

        :return: overall writing duration
        :rtype: float
        """
        return self.compute(self.wrapper, writing_duration_overall)

    def ratio_of_writing_durations(self):
        """
        Extracts the ratio of writing durations: on-surface / in-air.

        :return: ratio of writing durations
        :rtype: float
        """
        return self.compute(self.wrapper, ratio_of_writing_durations)

    def number_of_interruptions(self):
        """
        Extracts the number of interruptions.

        :return: number of interruptions
        :rtype: float
        """
        return self.compute(self.wrapper, number_of_interruptions)

    def number_of_interruptions_relative(self):
        """
        Extracts the number of interruptions relative to the duration.

        :return: number of interruptions relative to the duration
        :rtype: float
        """
        return self.compute(self.wrapper, number_of_interruptions_relative)

    # ---------------------
    # 5. Composite features

    def writing_tempo(self, in_air=False):
        """
        Extracts the writing tempo.

        :param in_air: in-air flag, defaults to False
        :type in_air: bool, optional
        :return: writing tempo
        :rtype: float
        """
        return self.compute(self.wrapper, writing_tempo, in_air=in_air)

    def writing_stops(self, statistics=()):
        """
        Extracts the writing stops.

        :param statistics: statistics to compute, defaults to ()
        :type statistics: Any[list, tuple], optional
        :return: writing stops
        :rtype: numpy.ndarray or np.NaN
        """
        return self.compute(self.wrapper, writing_stops, statistics=statistics)

    def number_of_changes_in_x_profile(self, fs, fc=None, n=None):
        """
        Extracts the number of changes in x profile.

        :param fs: sampling frequency
        :type fs: float
        :param fc: cutoff frequency for the low-pass filter, defaults to None
        :type fc: float, optional
        :param n: number of samples of a Gaussian filter, defaults to None
        :type n: int, optional
        :return: number of changes in x profile
        :rtype: int
        """
        return self.compute(self.wrapper, number_of_changes_in_x_profile, fs=fs, fc=fc, n=n)

    def number_of_changes_in_y_profile(self, fs, fc=None, n=None):
        """
        Extracts the number of changes in y profile.

        :param fs: sampling frequency
        :type fs: float
        :param fc: cutoff frequency for the low-pass filter, defaults to None
        :type fc: float, optional
        :param n: number of samples of a Gaussian filter, defaults to None
        :type n: int, optional
        :return: number of changes in y profile
        :rtype: int
        """
        return self.compute(self.wrapper, number_of_changes_in_y_profile, fs=fs, fc=fc, n=n)

    def number_of_changes_in_azimuth(self, fs, fc=None, n=None):
        """
        Extracts the number of changes in azimuth.

        :param fs: sampling frequency
        :type fs: float
        :param fc: cutoff frequency for the low-pass filter, defaults to None
        :type fc: float, optional
        :param n: number of samples of a Gaussian filter, defaults to None
        :type n: int, optional
        :return: number of changes in azimuth
        :rtype: int
        """
        return self.compute(self.wrapper, number_of_changes_in_azimuth, fs=fs, fc=fc, n=n)

    def number_of_changes_in_tilt(self, fs, fc=None, n=None):
        """
        Extracts the number of changes in tilt.

        :param fs: sampling frequency
        :type fs: float
        :param fc: cutoff frequency for the low-pass filter, defaults to None
        :type fc: float, optional
        :param n: number of samples of a Gaussian filter, defaults to None
        :type n: int, optional
        :return: number of changes in tilt
        :rtype: int
        """
        return self.compute(self.wrapper, number_of_changes_in_tilt, fs=fs, fc=fc, n=n)

    def number_of_changes_in_pressure(self, fs, fc=None, n=None):
        """
        Extracts the number of changes in pressure.

        :param fs: sampling frequency
        :type fs: float
        :param fc: cutoff frequency for the low-pass filter, defaults to None
        :type fc: float, optional
        :param n: number of samples of a Gaussian filter, defaults to None
        :type n: int, optional
        :return: number of changes in pressure
        :rtype: int
        """
        return self.compute(self.wrapper, number_of_changes_in_pressure, fs=fs, fc=fc, n=n)

    def number_of_changes_in_velocity_profile(self, fs, fc=None, n=None):
        """
        Extracts the number of changes in velocity profile.

        :param fs: sampling frequency
        :type fs: float
        :param fc: cutoff frequency for the low-pass filter, defaults to None
        :type fc: float, optional
        :param n: number of samples of a Gaussian filter, defaults to None
        :type n: int, optional
        :return: number of changes in velocity profile
        :rtype: int
        """
        return self.compute(self.wrapper, number_of_changes_in_velocity_profile, fs=fs, fc=fc, n=n)

    def relative_number_of_changes_in_x_profile(self, fs, fc=None, n=None):
        """
        Extracts the relative number of changes in x profile.

        :param fs: sampling frequency
        :type fs: float
        :param fc: cutoff frequency for the low-pass filter, defaults to None
        :type fc: float, optional
        :param n: number of samples of a Gaussian filter, defaults to None
        :type n: int, optional
        :return: relative number of changes in x profile
        :rtype: int
        """
        return self.compute(self.wrapper, relative_number_of_changes_in_x_profile, fs=fs, fc=fc, n=n)

    def relative_number_of_changes_in_y_profile(self, fs, fc=None, n=None):
        """
        Extracts the relative number of changes in y profile.

        :param fs: sampling frequency
        :type fs: float
        :param fc: cutoff frequency for the low-pass filter, defaults to None
        :type fc: float, optional
        :param n: number of samples of a Gaussian filter, defaults to None
        :type n: int, optional
        :return: relative number of changes in y profile
        :rtype: int
        """
        return self.compute(self.wrapper, relative_number_of_changes_in_y_profile, fs=fs, fc=fc, n=n)

    def relative_number_of_changes_in_azimuth(self, fs, fc=None, n=None):
        """
        Extracts the relative number of changes in azimuth.

        :param fs: sampling frequency
        :type fs: float
        :param fc: cutoff frequency for the low-pass filter, defaults to None
        :type fc: float, optional
        :param n: number of samples of a Gaussian filter, defaults to None
        :type n: int, optional
        :return: relative number of changes in azimuth
        :rtype: int
        """
        return self.compute(self.wrapper, relative_number_of_changes_in_azimuth, fs=fs, fc=fc, n=n)

    def relative_number_of_changes_in_tilt(self, fs, fc=None, n=None):
        """
        Extracts the relative number of changes in tilt.

        :param fs: sampling frequency
        :type fs: float
        :param fc: cutoff frequency for the low-pass filter, defaults to None
        :type fc: float, optional
        :param n: number of samples of a Gaussian filter, defaults to None
        :type n: int, optional
        :return: relative number of changes in tilt
        :rtype: int
        """
        return self.compute(self.wrapper, relative_number_of_changes_in_tilt, fs=fs, fc=fc, n=n)

    def relative_number_of_changes_in_pressure(self, fs, fc=None, n=None):
        """
        Extracts the relative number of changes in pressure.

        :param fs: sampling frequency
        :type fs: float
        :param fc: cutoff frequency for the low-pass filter, defaults to None
        :type fc: float, optional
        :param n: number of samples of a Gaussian filter, defaults to None
        :type n: int, optional
        :return: relative number of changes in pressure
        :rtype: int
        """
        return self.compute(self.wrapper, relative_number_of_changes_in_pressure, fs=fs, fc=fc, n=n)

    def relative_number_of_changes_in_velocity_profile(self, fs, fc=None, n=None):
        """
        Extracts the relative number of changes in velocity profile.

        :param fs: sampling frequency
        :type fs: float
        :param fc: cutoff frequency for the low-pass filter, defaults to None
        :type fc: float, optional
        :param n: number of samples of a Gaussian filter, defaults to None
        :type n: int, optional
        :return: relative number of changes in velocity profile
        :rtype: int
        """
        return self.compute(self.wrapper, relative_number_of_changes_in_velocity_profile, fs=fs, fc=fc, n=n)



from handwriting_features.features.exceptions.validation import *
from handwriting_features.features.configuration.settings import HandwritingFeaturesSettings


class HandwritingFeaturesFusion(object):
    """Class implementing the handwriting features fusion"""

    # Handwriting features settings
    features_settings = HandwritingFeaturesSettings.settings

    @classmethod
    def fuze(cls, feature_name, feature_args=None, common_configuration=None, skip_features=()):
        """
        Fuze the feature arguments with the common configuration.

        :param feature_name: feature name
        :type feature_name: str
        :param feature_args: feature arguments, defaults to None
        :type feature_args: dict, optional
        :param common_configuration: common configuration, defaults to None
        :type common_configuration: dict, optional
        :param skip_features: skip preparation for args names (features), defaults to ()
        :type skip_features: Any[list, tuple], optional
        :return: updated feature arguments
        :rtype: dict
        """

        # Prepare default feature args
        feature_args = feature_args if feature_args else {}

        # Get the features in the feature's configuration (to be used and configured)
        features_in_configuration = HandwritingFeaturesSettings.get_feature_arguments(feature_name).keys()

        # Update the feature arguments with the common configuration
        if common_configuration:
            for key, value in common_configuration.items():
                if key in feature_args:
                    continue
                if key in skip_features:
                    continue
                if value is None:
                    continue
                if key in features_in_configuration:
                    feature_args[key] = value

        # Return the updated feature arguments
        return feature_args


class HandwritingFeaturesValidation(object):
    """Class implementing the handwriting features validation"""

    # Handwriting features settings
    features_settings = HandwritingFeaturesSettings.settings

    @classmethod
    def validate(cls, feature_name, feature_args=None, skip_features=()):
        """
        Validate the feature arguments.

        :param feature_name: feature name
        :type feature_name: str
        :param feature_args: feature arguments, defaults to None
        :type feature_args: dict, optional
        :param skip_features: skip validation for args names (features), defaults to ()
        :type skip_features: Any[list, tuple], optional
        :return: validated feature arguments
        :rtype: dict
        """

        # Check the feature
        #
        # 1. check if the feature name is provided
        # 2. check if the type of feature name
        # 3. check if the feature name is known
        # 4. check if there are any arguments for the feature
        if not feature_name:
            raise FeatureNameMissingError("Missing feature name")
        if not isinstance(feature_name, str):
            raise FeatureNameInvalidTypeError(f"Unsupported feature type: {feature_name}")
        if feature_name not in cls.features_settings:
            raise FeatureNameUnsupportedError(f"Unsupported feature: {feature_name}")
        if not cls.features_settings.get(feature_name):
            return {}

        # Prepare default feature args
        feature_args = feature_args if feature_args else {}

        # Prepare the validated arguments
        validated_args = {}

        # ------------------------------- #
        # Validate the feature properties #
        # ------------------------------- #

        # 1. Validate the applicability of the statistical functions
        if not cls.features_settings.get(feature_name).get("properties").get("is_multi_valued"):
            if feature_args.get("statistics") and feature_args["statistics"]:
                raise StatisticsForSingleValuedFeatureError(
                    f"No statistics supported for single-valued features. "
                    f"feature: {feature_name}")

        # ------------------------------ #
        # Validate the feature arguments #
        # ------------------------------ #

        for argument_name, argument_settings in cls.features_settings.get(feature_name).get("arguments").items():

            # Handle skip validation
            if skip_features and argument_name in skip_features:
                continue

            # Get the argument data and type
            arg_data = feature_args.get(argument_name)
            arg_type = type(arg_data)

            # If the feature is optional, try to use a default value (if available)
            if not argument_settings.get("mandatory"):
                if arg_data is None and argument_settings.get("default"):
                    validated_args[argument_name] = argument_settings.get("default")

            # 1. Validate the presence of the mandatory argument
            if argument_name not in feature_args:
                if argument_settings.get("mandatory"):
                    raise FeatureArgumentMissingError(
                        f"Missing argument mandatory feature argument. "
                        f"argument name: {argument_name}, "
                        f"feature: {feature_name}")
                else:
                    continue

            # 2. Validate the type of the argument
            if argument_settings.get("type"):
                if arg_type not in argument_settings.get("type"):
                    raise FeatureArgumentInvalidTypeError(
                        f"Unsupported feature argument type: {arg_type}. "
                        f"argument name: {argument_name}, "
                        f"feature: {feature_name}")

            # 3. Validate the options of the argument
            if argument_settings.get("options"):
                if arg_data:
                    if not isinstance(arg_data, (list, tuple)):
                        arg_iter = [arg_data]
                    else:
                        arg_iter = arg_data

                    for i in arg_iter:
                        if i not in argument_settings.get("options"):
                            raise FeatureArgumentUnsupportedValueError(
                                f"Unsupported feature argument value: {i}. "
                                f"argument name: {argument_name}, "
                                f"feature: {feature_name}")

            # Add the validated argument
            validated_args[argument_name] = feature_args[argument_name]

        # Map the validated feature arguments
        return validated_args



from handwriting_features.data.containers.sample import HandwritingSampleWrapper
from handwriting_features.data.descriptors.statistics import Statistics


class HandwritingFeaturesSettings(object):
    """Class implementing the handwriting features settings"""

    # Handwriting data axes
    axes = HandwritingSampleWrapper.axes

    # Handwriting features statistics
    statistics = Statistics.mapping.keys()

    # Handwriting features settings
    settings = {

        # ---------------------
        # 1. Kinematic features

        # Velocity
        "velocity": {
            "properties": {
                "is_multi_valued": True
            },
            "arguments": {
                "axis": {
                    "mandatory": False,
                    "type": [str],
                    "options": axes,
                    "default": "xy"
                },
                "in_air": {
                    "mandatory": False,
                    "type": [bool],
                    "options": (True, False),
                    "default": False
                },
                "statistics": {
                    "mandatory": False,
                    "type": [str, list, tuple],
                    "options": statistics
                }
            }
        },

        # Acceleration
        "acceleration": {
            "properties": {
                "is_multi_valued": True
            },
            "arguments": {
                "axis": {
                    "mandatory": False,
                    "type": [str],
                    "options": axes,
                    "default": "xy"
                },
                "in_air": {
                    "mandatory": False,
                    "type": [bool],
                    "options": (True, False),
                    "default": False
                },
                "statistics": {
                    "mandatory": False,
                    "type": [str, list, tuple],
                    "options": statistics
                }
            }
        },

        # Jerk
        "jerk": {
            "properties": {
                "is_multi_valued": True
            },
            "arguments": {
                "axis": {
                    "mandatory": False,
                    "type": [str],
                    "options": axes,
                    "default": "xy"
                },
                "in_air": {
                    "mandatory": False,
                    "type": [bool],
                    "options": (True, False),
                    "default": False
                },
                "statistics": {
                    "mandatory": False,
                    "type": [str, list, tuple],
                    "options": statistics
                }
            }
        },

        # -------------------
        # 2. Dynamic features

        # Azimuth
        "azimuth": {
            "properties": {
                "is_multi_valued": True
            },
            "arguments": {
                "in_air": {
                    "mandatory": False,
                    "type": [bool],
                    "options": (True, False),
                    "default": False
                },
                "statistics": {
                    "mandatory": False,
                    "type": [str, list, tuple],
                    "options": statistics
                }
            }
        },

        # Tilt
        "tilt": {
            "properties": {
                "is_multi_valued": True
            },
            "arguments": {
                "in_air": {
                    "mandatory": False,
                    "type": [bool],
                    "options": (True, False),
                    "default": False
                },
                "statistics": {
                    "mandatory": False,
                    "type": [str, list, tuple],
                    "options": statistics
                }
            }
        },

        # Pressure
        "pressure": {
            "properties": {
                "is_multi_valued": True
            },
            "arguments": {
                "statistics": {
                    "mandatory": False,
                    "type": [str, list, tuple],
                    "options": statistics
                }
            }
        },

        # -------------------
        # 3. Spatial features

        # Stroke length
        "stroke_length": {
            "properties": {
                "is_multi_valued": True
            },
            "arguments": {
                "in_air": {
                    "mandatory": False,
                    "type": [bool],
                    "options": (True, False),
                    "default": False
                },
                "statistics": {
                    "mandatory": False,
                    "type": [str, list, tuple],
                    "options": statistics
                }
            }
        },

        # Stroke height
        "stroke_height": {
            "properties": {
                "is_multi_valued": True
            },
            "arguments": {
                "in_air": {
                    "mandatory": False,
                    "type": [bool],
                    "options": (True, False),
                    "default": False
                },
                "statistics": {
                    "mandatory": False,
                    "type": [str, list, tuple],
                    "options": statistics
                }
            }
        },

        # Stroke width
        "stroke_width": {
            "properties": {
                "is_multi_valued": True
            },
            "arguments": {
                "in_air": {
                    "mandatory": False,
                    "type": [bool],
                    "options": (True, False),
                    "default": False
                },
                "statistics": {
                    "mandatory": False,
                    "type": [str, list, tuple],
                    "options": statistics
                }
            }
        },

        # Writing length
        "writing_length": {
            "properties": {
                "is_multi_valued": False
            },
            "arguments": {
                "in_air": {
                    "mandatory": False,
                    "type": [bool],
                    "options": (True, False),
                    "default": False
                }
            }
        },

        # Writing height
        "writing_height": {
            "properties": {
                "is_multi_valued": False
            },
            "arguments": {
                "in_air": {
                    "mandatory": False,
                    "type": [bool],
                    "options": (True, False),
                    "default": False
                }
            }
        },

        # Writing width
        "writing_width": {
            "properties": {
                "is_multi_valued": False
            },
            "arguments": {
                "in_air": {
                    "mandatory": False,
                    "type": [bool],
                    "options": (True, False),
                    "default": False
                }
            }
        },

        # Number of intra-stroke intersections
        "number_of_intra_stroke_intersections": {
            "properties": {
                "is_multi_valued": True
            },
            "arguments": {
                "statistics": {
                    "mandatory": False,
                    "type": [str, list, tuple],
                    "options": statistics
                }
            }
        },

        # Relative number of intra-stroke intersections
        "relative_number_of_intra_stroke_intersections": {
            "properties": {
                "is_multi_valued": True
            },
            "arguments": {
                "statistics": {
                    "mandatory": False,
                    "type": [str, list, tuple],
                    "options": statistics
                }
            }
        },

        # Total number of intra-stroke intersections
        "total_number_of_intra_stroke_intersections": {
            "properties": {
                "is_multi_valued": False
            },
            "arguments": {}
        },

        # Relative total number of intra-stroke intersections
        "relative_total_number_of_intra_stroke_intersections": {
            "properties": {
                "is_multi_valued": False
            },
            "arguments": {}
        },

        # Number of inter-stroke intersections
        "number_of_inter_stroke_intersections": {
            "properties": {
                "is_multi_valued": False
            },
            "arguments": {}
        },

        # Relative number of inter-stroke intersections
        "relative_number_of_inter_stroke_intersections": {
            "properties": {
                "is_multi_valued": False
            },
            "arguments": {}
        },

        # Vertical peaks indices
        "vertical_peaks_indices": {
            "properties": {
                "is_multi_valued": True
            },
            "arguments": {
                "fs": {
                    "mandatory": True,
                    "type": [int, float]
                },
                "statistics": {
                    "mandatory": False,
                    "type": [str, list, tuple],
                    "options": statistics
                }
            }
        },

        # Vertical valleys indices
        "vertical_valleys_indices": {
            "properties": {
                "is_multi_valued": True
            },
            "arguments": {
                "fs": {
                    "mandatory": True,
                    "type": [int, float]
                },
                "statistics": {
                    "mandatory": False,
                    "type": [str, list, tuple],
                    "options": statistics
                }
            }
        },

        # Vertical peaks values
        "vertical_peaks_values": {
            "properties": {
                "is_multi_valued": True
            },
            "arguments": {
                "fs": {
                    "mandatory": True,
                    "type": [int, float]
                },
                "statistics": {
                    "mandatory": False,
                    "type": [str, list, tuple],
                    "options": statistics
                }
            }
        },

        # Vertical valleys values
        "vertical_valleys_values": {
            "properties": {
                "is_multi_valued": True
            },
            "arguments": {
                "fs": {
                    "mandatory": True,
                    "type": [int, float]
                },
                "statistics": {
                    "mandatory": False,
                    "type": [str, list, tuple],
                    "options": statistics
                }
            }
        },

        # Vertical peaks velocity
        "vertical_peaks_velocity": {
            "properties": {
                "is_multi_valued": True
            },
            "arguments": {
                "fs": {
                    "mandatory": True,
                    "type": [int, float]
                },
                "statistics": {
                    "mandatory": False,
                    "type": [str, list, tuple],
                    "options": statistics
                }
            }
        },

        # Vertical valleys velocity
        "vertical_valleys_velocity": {
            "properties": {
                "is_multi_valued": True
            },
            "arguments": {
                "fs": {
                    "mandatory": True,
                    "type": [int, float]
                },
                "statistics": {
                    "mandatory": False,
                    "type": [str, list, tuple],
                    "options": statistics
                }
            }
        },

        # Vertical peaks distance
        "vertical_peaks_distance": {
            "properties": {
                "is_multi_valued": True
            },
            "arguments": {
                "fs": {
                    "mandatory": True,
                    "type": [int, float]
                },
                "statistics": {
                    "mandatory": False,
                    "type": [str, list, tuple],
                    "options": statistics
                }
            }
        },

        # Vertical valleys distance
        "vertical_valleys_distance": {
            "properties": {
                "is_multi_valued": True
            },
            "arguments": {
                "fs": {
                    "mandatory": True,
                    "type": [int, float]
                },
                "statistics": {
                    "mandatory": False,
                    "type": [str, list, tuple],
                    "options": statistics
                }
            }
        },

        # Vertical peaks duration
        "vertical_peaks_duration": {
            "properties": {
                "is_multi_valued": True
            },
            "arguments": {
                "fs": {
                    "mandatory": True,
                    "type": [int, float]
                },
                "statistics": {
                    "mandatory": False,
                    "type": [str, list, tuple],
                    "options": statistics
                }
            }
        },

        # Vertical valleys duration
        "vertical_valleys_duration": {
            "properties": {
                "is_multi_valued": True
            },
            "arguments": {
                "fs": {
                    "mandatory": True,
                    "type": [int, float]
                },
                "statistics": {
                    "mandatory": False,
                    "type": [str, list, tuple],
                    "options": statistics
                }
            }
        },

        # --------------------
        # 4. Temporal features

        # Stroke duration
        "stroke_duration": {
            "properties": {
                "is_multi_valued": True
            },
            "arguments": {
                "in_air": {
                    "mandatory": False,
                    "type": [bool],
                    "options": (True, False),
                    "default": False
                },
                "statistics": {
                    "mandatory": False,
                    "type": [str, list, tuple],
                    "options": statistics
                }
            }
        },

        # Ratio of stroke durations
        "ratio_of_stroke_durations": {
            "properties": {
                "is_multi_valued": True
            },
            "arguments": {
                "statistics": {
                    "mandatory": False,
                    "type": [str, list, tuple],
                    "options": statistics
                }
            }
        },

        # Writing duration
        "writing_duration": {
            "properties": {
                "is_multi_valued": False
            },
            "arguments": {
                "in_air": {
                    "mandatory": False,
                    "type": [bool],
                    "options": (True, False),
                    "default": False
                }
            }
        },

        # Writing duration overall
        "writing_duration_overall": {
            "properties": {
                "is_multi_valued": False
            },
            "arguments": {}
        },

        # Ratio of writing durations
        "ratio_of_writing_durations": {
            "properties": {
                "is_multi_valued": False
            },
            "arguments": {}
        },

        # Number of interruptions
        "number_of_interruptions": {
            "properties": {
                "is_multi_valued": False
            },
            "arguments": {}
        },

        # Number of interruptions relative to the duration
        "number_of_interruptions_relative": {
            "properties": {
                "is_multi_valued": False
            },
            "arguments": {}
        },

        # ---------------------
        # 5. Composite features

        # Writing tempo
        "writing_tempo": {
            "properties": {
                "is_multi_valued": False
            },
            "arguments": {
                "in_air": {
                    "mandatory": False,
                    "type": [bool],
                    "options": (True, False),
                    "default": False
                }
            }
        },

        # Writing stops
        "writing_stops": {
            "properties": {
                "is_multi_valued": True
            },
            "arguments": {
                "statistics": {
                    "mandatory": False,
                    "type": [str, list, tuple],
                    "options": statistics
                }
            }
        },

        # Number of changes in x profile
        "number_of_changes_in_x_profile": {
            "properties": {
                "is_multi_valued": False
            },
            "arguments": {
                "fs": {
                    "mandatory": True,
                    "type": [int, float]
                }
            }
        },

        # Number of changes in y profile
        "number_of_changes_in_y_profile": {
            "properties": {
                "is_multi_valued": False
            },
            "arguments": {
                "fs": {
                    "mandatory": True,
                    "type": [int, float]
                }
            }
        },

        # Number of changes in azimuth
        "number_of_changes_in_azimuth": {
            "properties": {
                "is_multi_valued": False
            },
            "arguments": {
                "fs": {
                    "mandatory": True,
                    "type": [int, float]
                }
            }
        },

        # Number of changes in tilt
        "number_of_changes_in_tilt": {
            "properties": {
                "is_multi_valued": False
            },
            "arguments": {
                "fs": {
                    "mandatory": True,
                    "type": [int, float]
                }
            }
        },

        # Number of changes in pressure
        "number_of_changes_in_pressure": {
            "properties": {
                "is_multi_valued": False
            },
            "arguments": {
                "fs": {
                    "mandatory": True,
                    "type": [int, float]
                }
            }
        },

        # Number of changes in velocity profile
        "number_of_changes_in_velocity_profile": {
            "properties": {
                "is_multi_valued": False
            },
            "arguments": {
                "fs": {
                    "mandatory": True,
                    "type": [int, float]
                }
            }
        },

        # Relative number of changes in x profile
        "relative_number_of_changes_in_x_profile": {
            "properties": {
                "is_multi_valued": False
            },
            "arguments": {
                "fs": {
                    "mandatory": True,
                    "type": [int, float]
                }
            }
        },

        # Relative number of changes in y profile
        "relative_number_of_changes_in_y_profile": {
            "properties": {
                "is_multi_valued": False
            },
            "arguments": {
                "fs": {
                    "mandatory": True,
                    "type": [int, float]
                }
            }
        },

        # Relative number of changes in azimuth
        "relative_number_of_changes_in_azimuth": {
            "properties": {
                "is_multi_valued": False
            },
            "arguments": {
                "fs": {
                    "mandatory": True,
                    "type": [int, float]
                }
            }
        },

        # Relative number of changes in tilt
        "relative_number_of_changes_in_tilt": {
            "properties": {
                "is_multi_valued": False
            },
            "arguments": {
                "fs": {
                    "mandatory": True,
                    "type": [int, float]
                }
            }
        },

        # Relative number of changes in pressure
        "relative_number_of_changes_in_pressure": {
            "properties": {
                "is_multi_valued": False
            },
            "arguments": {
                "fs": {
                    "mandatory": True,
                    "type": [int, float]
                }
            }
        },

        # Relative number of changes in velocity profile
        "relative_number_of_changes_in_velocity_profile": {
            "properties": {
                "is_multi_valued": False
            },
            "arguments": {
                "fs": {
                    "mandatory": True,
                    "type": [int, float]
                }
            }
        }
    }

    @classmethod
    def get_feature_arguments(cls, feature_name):
        """
        Gets the feature's arguments.

        :param feature_name: feature name
        :type feature_name: str
        :return: arguments
        :rtype: iterable
        """
        return cls.settings.get(feature_name, {}).get("arguments", {})

    @classmethod
    def get_feature_argument_type(cls, feature_name, argument_name):
        """
        Gets the feature argument's supported type(s).

        :param feature_name: feature name
        :type feature_name: str
        :param argument_name: argument name
        :type argument_name: str
        :return: argument's type(s)
        :rtype: iterable
        """
        return cls.settings.get(feature_name, {}).get("arguments").get(argument_name, {}).get("type", [])

    @classmethod
    def get_feature_argument_options(cls, feature_name, argument_name):
        """
        Gets the feature argument's option(s).

        :param feature_name: feature name
        :type feature_name: str
        :param argument_name: argument name
        :type argument_name: str
        :return: argument's option(s)
        :rtype: iterable
        """
        return cls.settings.get(feature_name, {}).get("arguments").get(argument_name, {}).get("options", [])

    @classmethod
    def get_feature_argument_default(cls, feature_name, argument_name):
        """
        Gets the feature argument's default value(s).

        :param feature_name: feature name
        :type feature_name: str
        :param argument_name: argument name
        :type argument_name: str
        :return: argument's default value(s)
        :rtype: Any
        """
        return cls.settings.get(feature_name, {}).get("arguments").get(argument_name, {}).get("default")

    @classmethod
    def is_feature_multivalued(cls, feature_name):
        """
        Checks if the feature is multivalued.

        :param feature_name: feature name
        :type feature_name: str
        :return: True if the feature is multivalued, False otherwise
        :rtype: bool
        """
        return cls.settings.get(feature_name, {}).get("properties", {}).get("is_multi_valued", False)



def azimuth(sample_wrapper, in_air):
    """
    Returns azimuth.

    :param sample_wrapper: sample wrapper object
    :type sample_wrapper: HandwritingSampleWrapper
    :param in_air: in-air flag
    :type in_air: bool
    :return: azimuth
    :rtype: numpy.ndarray or np.NaN
    """
    return sample_wrapper.compute_azimuth(in_air)


def tilt(sample_wrapper, in_air):
    """
    Returns tilt.

    :param sample_wrapper: sample wrapper object
    :type sample_wrapper: HandwritingSampleWrapper
    :param in_air: in-air flag
    :type in_air: bool
    :return: tilt
    :rtype: numpy.ndarray or np.NaN
    """
    return sample_wrapper.compute_tilt(in_air)


def pressure(sample_wrapper):
    """
    Returns pressure.

    :param sample_wrapper: sample wrapper object
    :type sample_wrapper: HandwritingSampleWrapper
    :return: pressure
    :rtype: numpy.ndarray or np.NaN
    """
    return sample_wrapper.compute_pressure()



class UnsupportedDataForStatisticsError(Exception):
    """Raised when unsupported data are used to compute statistics"""
    pass


class StatisticsNameNotInMappingError(Exception):
    """Raised when a statistics name is not in the statistical mapping"""
    pass



import numpy
import pandas


def derivation(array, order=1):
    """
    Returns a derivation of input <array>.

    :param array: input array
    :type array: numpy.ndarray
    :param order: order of derivation, defaults to 1
    :type order: int, optional
    :return: derivation
    :rtype: numpy.ndarray
    """
    return numpy.diff(array, order, axis=0) if isinstance(array, numpy.ndarray) else numpy.nan


def intersection(x1, y1, x2=None, y2=None):
    """
    Computes the intersection of two curves.
    Inspired by: https://www.mathworks.com/matlabcentral/fileexchange/22441-curve-intersections

    :param x1: x-values of the first curve
    :type x1: numpy.ndarray
    :param y1: y-values of the first curve
    :type y1: numpy.ndarray
    :param x2: x-values of the second curve
    :type x2: numpy.ndarray
    :param y2: y-values of the second curve
    :type y2: numpy.ndarray
    :return: intersections
    :rtype: numpy.ndarray
    """

    # Prepare the input variables
    x1 = pandas.DataFrame(x1) if not isinstance(x1, pandas.DataFrame) else x1
    y1 = pandas.DataFrame(y1) if not isinstance(y1, pandas.DataFrame) else y1
    x2 = pandas.DataFrame(x2) if (x2 is not None and not isinstance(x2, pandas.DataFrame)) else x2
    y2 = pandas.DataFrame(y2) if (y2 is not None and not isinstance(y2, pandas.DataFrame)) else y2

    # Prepare the handling of the intersections (based on the presence of the second curve)
    #
    # 1. if the second curve is not provided, shared points will not be returned
    # 2. if the second curve is provided, all intersection points will be returned
    second_curve_exists = False if x2 is None else True

    # Handle the presence of the second curve
    if not second_curve_exists:
        x2 = x1
        y2 = y1
        x2 = numpy.mat(x2.values).transpose()
        y2 = numpy.mat(y2.values).transpose()
    else:
        x2 = numpy.mat(x2.values).transpose()
        y2 = numpy.mat(y2.values).transpose()

    # Convert the 1st curve to column-vectors
    x1 = numpy.mat(x1.values)
    y1 = numpy.mat(y1.values)

    # Combine x and y for each curve to one master matrix
    #
    # 1. the 1st curve - column vector
    # 2. the 2nd curve - row vector
    curve1 = numpy.hstack([x1, y1])
    curve2 = numpy.vstack([x2, y2])

    x1 = curve1.transpose()[0].transpose()
    y1 = curve1.transpose()[1].transpose()
    x2 = curve2[0]
    y2 = curve2[1]

    # Compute the distance between adjacent x's and y's
    dx1 = numpy.diff(x1, axis=0)
    dx2 = numpy.diff(x2)
    dy1 = numpy.diff(y1, axis=0)
    dy2 = numpy.diff(y2)

    # Compute the signed differences
    s1 = numpy.multiply(dx1, y1[0:len(y1) - 1]) - numpy.multiply(dy1, x1[0:len(x1) - 1])
    s2 = \
        numpy.multiply(dx2, y2.transpose()[0:len(y2.transpose()) - 1].transpose()) - \
        numpy.multiply(dy2, x2.transpose()[0:len(x2.transpose()) - 1].transpose())

    f1 = dx1 * y2 - dy1 * x2
    f2 = y1 * dx2 - x1 * dy2

    f2t = f2.transpose()
    s2t = s2.transpose()

    # Collected distances between points in one curve and line segments in the other
    c1 = numpy.multiply(
        f1[0:int(f1.shape[0]), 0:int(f1.shape[1]) - 1] - s1,
        f1[0:int(f1.shape[0]), 1:int(f1.shape[1])] - s1)
    c2 = numpy.multiply(
        f2t[0:int(f2t.shape[0]), 0:int(f2t.shape[1]) - 1] - s2t,
        f2t[0:int(f2t.shape[0]), 1:int(f2t.shape[1])] - s2t)

    # If looking for self-intersections, take only points that aren't tangential
    if not second_curve_exists:
        tf1 = c1 < 0
        tf2 = c2 < 0

    # Otherwise, take tangent points as well
    else:
        tf1 = c1 <= 0
        tf2 = c2 <= 0

    tf2 = tf2.transpose()

    # Keep indicates row and column indices of line segments where intersections between the two curves are expected
    keep = tf1 & tf2

    # Collect row and column index values from the keep matrix
    i = []
    j = []
    for row in range(keep.shape[0]):
        for column in range(keep.shape[1]):
            if keep[row, column]:
                i.append(row)
                j.append(column)

    if not i:
        return pandas.DataFrame({"xs": [None], "xy": [None]}, columns=["xs", "ys"])

    # Transpose the distances
    dy2 = dy2.transpose()
    dx2 = dx2.transpose()

    # Prepare the computation
    r = numpy.multiply(dy2[j], dx1[i]) - numpy.multiply(dy1[i], dx2[j])

    r_prime = []
    i_prime = []
    j_prime = []

    for num in range(r.shape[0]):
        if r[num] != 0:
            r_prime.append(r[num, 0])
            i_prime.append(i[num])
            j_prime.append(j[num])

    r_prime = numpy.mat(r_prime).transpose()

    # Set up numerator and denominator to solve the system of equations
    numerator = numpy.hstack([
        numpy.multiply(dx2[j_prime], s1[i_prime]) - numpy.multiply(dx1[i_prime], s2t[j_prime]),
        numpy.multiply(dy2[j_prime], s1[i_prime]) - numpy.multiply(dy1[i_prime], s2t[j_prime])
    ])
    denominator = numpy.hstack([
        r_prime,
        r_prime
    ])

    # Solve the system of equations
    result = numpy.divide(numerator, denominator)

    # Organize the intersections into pandas DataFrame
    intersections = pandas.DataFrame(result, columns=["xs", "ys"])
    intersections = intersections.drop_duplicates(inplace=False)

    # Return the intersections
    return intersections.values



import numpy
import functools
from copy import deepcopy
from handwriting_features.data.utils.math import derivation
from handwriting_features.data.utils.dsp import LowPassFilter, GaussianFilter
from handwriting_features.features.implementation.conventional.spatial import stroke_length
from handwriting_features.features.implementation.conventional.temporal import stroke_duration


class WritingTempoUtils(object):
    """Class implementing writing tempo utils"""

    def __init__(self, sample_wrapper, in_air):
        """Initializes the writing tempo utils object"""

        # Set the stroke lengths and durations
        self.stroke_lengths = stroke_length(sample_wrapper, in_air)
        self.stroke_durations = stroke_duration(sample_wrapper, in_air)

    def get_writing_tempo(self):
        """Extracts the writing tempo"""
        return len(self.stroke_lengths) / (sum(self.stroke_durations) + numpy.finfo(float).eps)


class WritingStopsUtils(object):
    """Class implementing writing stops utils"""

    def __init__(self, sample_wrapper):
        """Initializes the writing stops utils object"""

        # Set the strokes and indices
        self.strokes = sample_wrapper.strokes
        self.indices = self._get_stroke_indexes([stroke for _, stroke in self.strokes])

    def get_writing_stops(self):
        """Extracts the stroke stops"""

        # Prepare the stops
        time = numpy.nan
        stops_borders_left = []
        stops_borders_right = []

        # Extract the stops
        for (pen_status, stroke), index in zip(self.strokes, self.indices):
            if pen_status == "in_air":
                continue

            # Compute the vector of length, time
            length = numpy.sqrt(numpy.power(derivation(stroke.x), 2) + numpy.power(derivation(stroke.y), 2))
            time = derivation(stroke.time)

            # Compute the vector of velocity (value <= 1 mm/s is set to 0)
            velocity = (d / t for (d, t) in zip(length, time))
            velocity = numpy.array([0 if v <= 1 else v for v in velocity])

            # Get the number of samples equaling to 15 ms
            num_samples = numpy.ceil(0.015 / numpy.mean(time))

            # Identify the stops
            border_left, border_right = self._get_borders(velocity)

            # Take only pauses lasting at least 15 ms
            pause_indices = numpy.where((border_right - border_left) > num_samples)[0]
            border_left = border_left[pause_indices].astype(float)
            border_right = border_right[pause_indices].astype(float)

            # Fuze the pauses
            border_left, border_right = self._fuze_pauses(border_left, border_right, num_samples)

            # Add the starting index of the stroke
            border_left += index
            border_right += index - 1

            # Append the borders to the stops
            stops_borders_left += border_left.tolist()
            stops_borders_right += border_right.tolist()

        # Get the writing stops
        stops = (numpy.array(stops_borders_right) - numpy.array(stops_borders_left)) * numpy.mean(time)
        stops = numpy.array(stops)

        # Return the writing stops
        return stops

    @classmethod
    def _get_stroke_indexes(cls, strokes):
        """Gets the indexes of the strokes"""

        # Prepare the list of indexes
        indexes = []

        # Get the indexes
        for i in range(len(strokes)):
            if i == 0:
                indexes.append(0)
            else:
                indexes.append(indexes[i - 1] + len(strokes[i - 1].time))

        # Return the indexes
        return indexes

    @classmethod
    def _get_borders(cls, array, border_value=0):
        """Gets borders of an array given a border value"""

        # Get the shifted arrays
        array_l = array[:-1]
        array_r = array[1:]

        # Get the borders
        border_l = numpy.where(numpy.logical_and(array_r == border_value, array_l != border_value))[0]
        border_r = numpy.where(numpy.logical_and(array_r != border_value, array_l == border_value))[0]

        if array[0] == border_value:
            border_l = numpy.array([0] + border_l.tolist())
        if array[-1] == border_value:
            border_r = numpy.array(border_r.tolist() + [len(array)])

        # Return the borders
        return border_l, border_r

    @classmethod
    def _fuze_pauses(cls, border_left, border_right, num_samples):
        """Fuzes the pauses"""

        # Fuze the pauses
        if len(border_left) > 1:
            for i in range(len(border_left) - 1):
                if border_left[i + 1] - border_right[i] < (2 * num_samples):
                    border_left[i + 1], border_right[i] = numpy.nan, numpy.nan

            # Get the improper pauses
            to_remove = [
                i for i, (l, r) in enumerate(zip(border_left, border_right))
                if numpy.isnan(l) and numpy.isnan(r)
            ]

            # Remove improper pauses
            border_left = numpy.delete(border_left, to_remove)
            border_right = numpy.delete(border_right, to_remove)

            # Update the pauses
            nans_right = numpy.isnan(border_right)
            if nans_right.any():

                border_right = numpy.array([
                    border_right[i + 1] if is_nan else border_right[i]
                    for i, is_nan in zip(range(len(border_right)), nans_right)
                ])

                border_left = numpy.delete(border_left, numpy.where(nans_right)[0] + 1)
                border_right = numpy.delete(border_right, numpy.where(nans_right)[0] + 1)

        # Return the fused pauses
        return border_left, border_right


class WritingNumberOfChangesUtils(object):
    """Class implementing writing number of changes utils"""

    # Default computational arguments
    fc = 17.5
    n = 50

    def __init__(self, sample_wrapper, fs, fc=None, n=None, subset=None):
        """
        Initializes the writing number of changes utils object.

        :param fs: sampling frequency
        :type fs: float
        :param fc: cutoff frequency for the low-pass filter, defaults to 17.5
        :type fc: float, optional
        :param n: number of samples of a Gaussian filter, defaults to 50
        :type n: int, optional
        :param subset: subset of composite features to return, defaults to None
        :type subset: list, optional
        """

        # Set the sample wrapper
        self.sample_wrapper = deepcopy(sample_wrapper)

        # Set the DSP arguments
        self.fs = fs
        self.fc = fc if fc else WritingNumberOfChangesUtils.fc
        self.n = n if n else WritingNumberOfChangesUtils.n

        # Set the filter instances
        self.low_pass_filter = LowPassFilter(self.fs, self.fc)
        self.gaussian_filter = GaussianFilter(self.fs, self.n)

        # Get the duration
        self.duration = self.sample_wrapper.sample_time[-1] - self.sample_wrapper.sample_time[0]

        # Filter x, y, azimuth, tilt, and pressure by a low-pass filter
        self.sample_wrapper.sample = self._filter_data_with_low_pass_filter(self.sample_wrapper.sample)

        # Get the on-surface strokes
        self.strokes = self.sample_wrapper.on_surface_strokes

        # Set the subset of the features to return
        self.subset = subset

    @functools.lru_cache(maxsize=1)
    def get_number_of_changes_in_x_profile(self):
        """Gets the number of changes in x profile"""
        return sum(self._get_changes(self._filter_data_with_gaussian_filter(s).x) for s in self.strokes)

    @functools.lru_cache(maxsize=1)
    def get_number_of_changes_in_y_profile(self):
        """Gets the number of changes in y profile"""
        return sum(self._get_changes(self._filter_data_with_gaussian_filter(s).y) for s in self.strokes)

    @functools.lru_cache(maxsize=1)
    def get_number_of_changes_in_azimuth(self):
        """Gets the number of changes in azimuth"""
        return sum(self._get_changes(self._filter_data_with_gaussian_filter(s).azimuth) for s in self.strokes)

    @functools.lru_cache(maxsize=1)
    def get_number_of_changes_in_tilt(self):
        """Gets the number of changes in tilt"""
        return sum(self._get_changes(self._filter_data_with_gaussian_filter(s).tilt) for s in self.strokes)

    @functools.lru_cache(maxsize=1)
    def get_number_of_changes_in_pressure(self):
        """Gets the number of changes in pressure"""
        return sum(self._get_changes(self._filter_data_with_gaussian_filter(s).pressure) for s in self.strokes)

    @functools.lru_cache(maxsize=1)
    def get_number_of_changes_in_velocity_profile(self):
        """Gets the number of changes in velocity profile"""

        # Prepare the number of changes
        num_changes = 0

        # Compute the number of changes
        for stroke in self.strokes:

            # Compute the vector of length, time, and velocity
            length = numpy.sqrt(numpy.power(derivation(stroke.x), 2) + numpy.power(derivation(stroke.y), 2))
            time = derivation(stroke.time)
            velocity = numpy.array([d / t for (d, t) in zip(length, time)])

            # Filter the velocity by a Gaussian filter
            velocity = self._filter_velocity_with_gaussian_filter(velocity)

            # Compute the number of changes
            num_changes += self._get_changes(velocity)

        # Return the number of changes
        return num_changes

    def get_relative_number_of_changes_in_x_profile(self):
        """Gets the relative number of changes in x profile"""
        return self.get_number_of_changes_in_x_profile() / self.duration

    def get_relative_number_of_changes_in_y_profile(self):
        """Gets the relative number of changes in y profile"""
        return self.get_number_of_changes_in_y_profile() / self.duration

    def get_relative_number_of_changes_in_azimuth(self):
        """Gets the relative number of changes in azimuth"""
        return self.get_number_of_changes_in_azimuth() / self.duration

    def get_relative_number_of_changes_in_tilt(self):
        """Gets the relative number of changes in tilt"""
        return self.get_number_of_changes_in_tilt() / self.duration

    def get_relative_number_of_changes_in_pressure(self):
        """Gets the relative number of changes in pressure"""
        return self.get_number_of_changes_in_pressure() / self.duration

    def get_relative_number_of_changes_in_velocity_profile(self):
        """Gets the relative number of changes in velocity profile"""
        return self.get_number_of_changes_in_velocity_profile() / self.duration

    @classmethod
    def _get_changes(cls, signal):
        """Gets the changes"""

        # Get the left/right side changes
        changes_left = numpy.sum(numpy.logical_and(signal[1:-1] > signal[:-2], signal[1:-1] > signal[2:]))
        changes_right = numpy.sum(numpy.logical_and(signal[1:-1] < signal[:-2], signal[1:-1] < signal[2:]))

        # Return the changes
        return changes_left + changes_right

    def _filter_data_with_low_pass_filter(self, signal):
        """Filters an input sample/stroke data by a low-pass filter"""

        # Filter the signal
        signal.x = self.low_pass_filter.filter(signal.x)
        signal.y = self.low_pass_filter.filter(signal.y)
        signal.azimuth = self.low_pass_filter.filter(signal.azimuth)
        signal.tilt = self.low_pass_filter.filter(signal.tilt)
        signal.pressure = self.low_pass_filter.filter(signal.pressure)

        # Return the filtered signal
        return signal

    def _filter_data_with_gaussian_filter(self, signal):
        """Filters an input sample/stroke data by a Gaussian filter"""

        # Filter the signal
        signal.x = self.gaussian_filter.filter(signal.x)
        signal.y = self.gaussian_filter.filter(signal.y)
        signal.azimuth = self.gaussian_filter.filter(signal.azimuth)
        signal.tilt = self.gaussian_filter.filter(signal.tilt)
        signal.pressure = self.gaussian_filter.filter(signal.pressure)

        # Return the filtered signal
        return signal

    def _filter_velocity_with_low_pass_filter(self, velocity):
        """Filters an input velocity by a low-pass filter"""
        return self.low_pass_filter.filter(velocity)

    def _filter_velocity_with_gaussian_filter(self, velocity):
        """Filters an input velocity by a Gaussian filter"""
        return self.gaussian_filter.filter(velocity)



# Handwriting sample-specific exceptions
from handwriting_sample.reader.exceptions import *
from handwriting_sample.transformer.exceptions import *
from handwriting_sample.validator.exceptions import *
from handwriting_sample.writer.exceptions import *

# Handwriting features-specific exceptions
from handwriting_features.data.exceptions.sample import *
from handwriting_features.data.exceptions.statistics import *
from handwriting_features.features.exceptions.mapping import *
from handwriting_features.features.exceptions.validation import *



import numpy


def velocity(sample_wrapper, axis, in_air):
    """
    Returns velocity.

    :param sample_wrapper: sample wrapper object
    :type sample_wrapper: HandwritingSampleWrapper
    :param axis: axis to compute the velocity from
    :type axis: str
    :param in_air: in-air flag
    :type in_air: bool
    :return: velocity
    :rtype: numpy.ndarray or np.NaN
    """
    return sample_wrapper.compute_velocity(axis, in_air)


def acceleration(sample_wrapper, axis, in_air):
    """
    Returns acceleration.

    :param sample_wrapper: sample wrapper object
    :type sample_wrapper: HandwritingSampleWrapper
    :param axis: axis to compute the acceleration from
    :type axis: str
    :param in_air: in-air flag
    :type in_air: bool
    :return: acceleration
    :rtype: numpy.ndarray or np.NaN
    """
    return sample_wrapper.compute_acceleration(axis, in_air)


def jerk(sample_wrapper, axis, in_air):
    """
    Returns jerk.

    :param sample_wrapper: sample wrapper object
    :type sample_wrapper: HandwritingSampleWrapper
    :param axis: axis to compute the jerk from
    :type axis: str
    :param in_air: in-air flag
    :type in_air: bool
    :return: jerk
    :rtype: numpy.ndarray or np.NaN
    """
    return sample_wrapper.compute_jerk(axis, in_air)



from handwriting_features.features.exceptions.mapping import *


class HandwritingFeaturesMapping(object):
    """Class implementing the handwriting features mapping"""

    def __init__(self, features):
        """Constructor method"""
        self.features = features

    def map(self, feature_name):
        """Map the feature name to the actual feature computation method"""

        # Prepare the mapping (feature name: feature computation method)
        mapping = {

            # 1. Kinematic features
            "velocity":
                self.features.velocity,
            "acceleration":
                self.features.acceleration,
            "jerk":
                self.features.jerk,

            # 2. Dynamic features
            "azimuth":
                self.features.azimuth,
            "tilt":
                self.features.tilt,
            "pressure":
                self.features.pressure,

            # 3. Spatial features
            "stroke_length":
                self.features.stroke_length,
            "stroke_height":
                self.features.stroke_height,
            "stroke_width":
                self.features.stroke_width,
            "writing_length":
                self.features.writing_length,
            "writing_height":
                self.features.writing_height,
            "writing_width":
                self.features.writing_width,
            "number_of_intra_stroke_intersections":
                self.features.number_of_intra_stroke_intersections,
            "relative_number_of_intra_stroke_intersections":
                self.features.relative_number_of_intra_stroke_intersections,
            "total_number_of_intra_stroke_intersections":
                self.features.total_number_of_intra_stroke_intersections,
            "relative_total_number_of_intra_stroke_intersections":
                self.features.relative_total_number_of_intra_stroke_intersections,
            "number_of_inter_stroke_intersections":
                self.features.number_of_inter_stroke_intersections,
            "relative_number_of_inter_stroke_intersections":
                self.features.relative_number_of_inter_stroke_intersections,
            "vertical_peaks_indices":
                self.features.vertical_peaks_indices,
            "vertical_valleys_indices":
                self.features.vertical_valleys_indices,
            "vertical_peaks_values":
                self.features.vertical_peaks_values,
            "vertical_valleys_values":
                self.features.vertical_valleys_values,
            "vertical_peaks_velocity":
                self.features.vertical_peaks_velocity,
            "vertical_valleys_velocity":
                self.features.vertical_valleys_velocity,
            "vertical_peaks_distance":
                self.features.vertical_peaks_distance,
            "vertical_valleys_distance":
                self.features.vertical_valleys_distance,
            "vertical_peaks_duration":
                self.features.vertical_peaks_duration,
            "vertical_valleys_duration":
                self.features.vertical_valleys_duration,

            # 4. Temporal features
            "stroke_duration":
                self.features.stroke_duration,
            "ratio_of_stroke_durations":
                self.features.ratio_of_stroke_durations,
            "writing_duration":
                self.features.writing_duration,
            "writing_duration_overall":
                self.features.writing_duration_overall,
            "ratio_of_writing_durations":
                self.features.ratio_of_writing_durations,
            "number_of_interruptions":
                self.features.number_of_interruptions,
            "number_of_interruptions_relative":
                self.features.number_of_interruptions_relative,

            # 5. Composite features
            "writing_tempo":
                self.features.writing_tempo,
            "writing_stops":
                self.features.writing_stops,
            "number_of_changes_in_x_profile":
                self.features.number_of_changes_in_x_profile,
            "number_of_changes_in_y_profile":
                self.features.number_of_changes_in_y_profile,
            "number_of_changes_in_azimuth":
                self.features.number_of_changes_in_azimuth,
            "number_of_changes_in_tilt":
                self.features.number_of_changes_in_tilt,
            "number_of_changes_in_pressure":
                self.features.number_of_changes_in_pressure,
            "number_of_changes_in_velocity_profile":
                self.features.number_of_changes_in_velocity_profile,
            "relative_number_of_changes_in_x_profile":
                self.features.relative_number_of_changes_in_x_profile,
            "relative_number_of_changes_in_y_profile":
                self.features.relative_number_of_changes_in_y_profile,
            "relative_number_of_changes_in_azimuth":
                self.features.relative_number_of_changes_in_azimuth,
            "relative_number_of_changes_in_tilt":
                self.features.relative_number_of_changes_in_tilt,
            "relative_number_of_changes_in_pressure":
                self.features.relative_number_of_changes_in_pressure,
            "relative_number_of_changes_in_velocity_profile":
                self.features.relative_number_of_changes_in_velocity_profile,
        }

        # Map the feature name to the feature computation method
        try:
            return mapping[feature_name]
        except KeyError:
            raise FeatureNameNotInMappingError(f"No mapping available for feature {feature_name}")



from handwriting_features.features.implementation.conventional.kinematic import *
from handwriting_features.features.implementation.conventional.dynamic import *
from handwriting_features.features.implementation.conventional.spatial import *
from handwriting_features.features.implementation.conventional.temporal import *
from handwriting_features.features.implementation.conventional.composite import *



from handwriting_features.features.implementation.conventional.utils.composite import (
    WritingTempoUtils,
    WritingStopsUtils,
    WritingNumberOfChangesUtils
)


def writing_tempo(sample_wrapper, in_air):
    """
    Returns writing tempo.

    :param sample_wrapper: sample wrapper object
    :type sample_wrapper: HandwritingSampleWrapper
    :param in_air: in-air flag
    :type in_air: bool
    :return: writing tempo
    :rtype: float
    """
    WritingTempoUtils(sample_wrapper, in_air).get_writing_tempo()


def writing_stops(sample_wrapper):
    """
    Returns writing stops.

    :param sample_wrapper: sample wrapper object
    :type sample_wrapper: HandwritingSampleWrapper
    :return: writing stops
    :rtype: numpy.ndarray or np.NaN
    """
    return WritingStopsUtils(sample_wrapper).get_writing_stops()


def number_of_changes_in_x_profile(sample_wrapper, fs, fc=None, n=None):
    """
    Returns number of changes in x profile.

    :param sample_wrapper: sample wrapper object
    :type sample_wrapper: HandwritingSampleWrapper
    :param fs: sampling frequency
    :type fs: float
    :param fc: cutoff frequency for the low-pass filter, defaults to None
    :type fc: float, optional
    :param n: number of samples of a Gaussian filter, defaults to None
    :type n: int, optional
    :return: number of changes in x profile
    :rtype: int
    """
    return WritingNumberOfChangesUtils(sample_wrapper, fs=fs, fc=fc, n=n).get_number_of_changes_in_x_profile()


def number_of_changes_in_y_profile(sample_wrapper, fs, fc=None, n=None):
    """
    Returns number of changes in y profile.

    :param sample_wrapper: sample wrapper object
    :type sample_wrapper: HandwritingSampleWrapper
    :param fs: sampling frequency
    :type fs: float
    :param fc: cutoff frequency for the low-pass filter, defaults to None
    :type fc: float, optional
    :param n: number of samples of a Gaussian filter, defaults to None
    :type n: int, optional
    :return: number of changes in y profile
    :rtype: int
    """
    return WritingNumberOfChangesUtils(sample_wrapper, fs=fs, fc=fc, n=n).get_number_of_changes_in_y_profile()


def number_of_changes_in_azimuth(sample_wrapper, fs, fc=None, n=None):
    """
    Returns number of changes in azimuth.

    :param sample_wrapper: sample wrapper object
    :type sample_wrapper: HandwritingSampleWrapper
    :param fs: sampling frequency
    :type fs: float
    :param fc: cutoff frequency for the low-pass filter, defaults to None
    :type fc: float, optional
    :param n: number of samples of a Gaussian filter, defaults to None
    :type n: int, optional
    :return: number of changes in azimuth
    :rtype: int
    """
    return WritingNumberOfChangesUtils(sample_wrapper, fs=fs, fc=fc, n=n).get_number_of_changes_in_azimuth()


def number_of_changes_in_tilt(sample_wrapper, fs, fc=None, n=None):
    """
    Returns number of changes in tilt.

    :param sample_wrapper: sample wrapper object
    :type sample_wrapper: HandwritingSampleWrapper
    :param fs: sampling frequency
    :type fs: float
    :param fc: cutoff frequency for the low-pass filter, defaults to None
    :type fc: float, optional
    :param n: number of samples of a Gaussian filter, defaults to None
    :type n: int, optional
    :return: number of changes in tilt
    :rtype: int
    """
    return WritingNumberOfChangesUtils(sample_wrapper, fs=fs, fc=fc, n=n).get_number_of_changes_in_tilt()


def number_of_changes_in_pressure(sample_wrapper, fs, fc=None, n=None):
    """
    Returns number of changes in pressure.

    :param sample_wrapper: sample wrapper object
    :type sample_wrapper: HandwritingSampleWrapper
    :param fs: sampling frequency
    :type fs: float
    :param fc: cutoff frequency for the low-pass filter, defaults to None
    :type fc: float, optional
    :param n: number of samples of a Gaussian filter, defaults to None
    :type n: int, optional
    :return: number of changes in pressure
    :rtype: int
    """
    return WritingNumberOfChangesUtils(sample_wrapper, fs=fs, fc=fc, n=n).get_number_of_changes_in_pressure()


def number_of_changes_in_velocity_profile(sample_wrapper, fs, fc=None, n=None):
    """
    Returns number of changes in velocity profile.

    :param sample_wrapper: sample wrapper object
    :type sample_wrapper: HandwritingSampleWrapper
    :param fs: sampling frequency
    :type fs: float
    :param fc: cutoff frequency for the low-pass filter, defaults to None
    :type fc: float, optional
    :param n: number of samples of a Gaussian filter, defaults to None
    :type n: int, optional
    :return: number of changes in velocity profile
    :rtype: int
    """
    return WritingNumberOfChangesUtils(
        sample_wrapper=sample_wrapper,
        fs=fs,
        fc=fc,
        n=n).get_number_of_changes_in_velocity_profile()


def relative_number_of_changes_in_x_profile(sample_wrapper, fs, fc=None, n=None):
    """
    Returns relative number of changes in x profile.

    :param sample_wrapper: sample wrapper object
    :type sample_wrapper: HandwritingSampleWrapper
    :param fs: sampling frequency
    :type fs: float
    :param fc: cutoff frequency for the low-pass filter, defaults to None
    :type fc: float, optional
    :param n: number of samples of a Gaussian filter, defaults to None
    :type n: int, optional
    :return: relative number of changes in x profile
    :rtype: int
    """
    return WritingNumberOfChangesUtils(sample_wrapper, fs=fs, fc=fc, n=n).get_relative_number_of_changes_in_x_profile()


def relative_number_of_changes_in_y_profile(sample_wrapper, fs, fc=None, n=None):
    """
    Returns relative number of changes in y profile.

    :param sample_wrapper: sample wrapper object
    :type sample_wrapper: HandwritingSampleWrapper
    :param fs: sampling frequency
    :type fs: float
    :param fc: cutoff frequency for the low-pass filter, defaults to None
    :type fc: float, optional
    :param n: number of samples of a Gaussian filter, defaults to None
    :type n: int, optional
    :return: relative number of changes in y profile
    :rtype: int
    """
    return WritingNumberOfChangesUtils(sample_wrapper, fs=fs, fc=fc, n=n).get_relative_number_of_changes_in_y_profile()


def relative_number_of_changes_in_azimuth(sample_wrapper, fs, fc=None, n=None):
    """
    Returns relative number of changes in azimuth.

    :param sample_wrapper: sample wrapper object
    :type sample_wrapper: HandwritingSampleWrapper
    :param fs: sampling frequency
    :type fs: float
    :param fc: cutoff frequency for the low-pass filter, defaults to None
    :type fc: float, optional
    :param n: number of samples of a Gaussian filter, defaults to None
    :type n: int, optional
    :return: relative number of changes in azimuth
    :rtype: int
    """
    return WritingNumberOfChangesUtils(sample_wrapper, fs=fs, fc=fc, n=n).get_relative_number_of_changes_in_azimuth()


def relative_number_of_changes_in_tilt(sample_wrapper, fs, fc=None, n=None):
    """
    Returns relative number of changes in tilt.

    :param sample_wrapper: sample wrapper object
    :type sample_wrapper: HandwritingSampleWrapper
    :param fs: sampling frequency
    :type fs: float
    :param fc: cutoff frequency for the low-pass filter, defaults to None
    :type fc: float, optional
    :param n: number of samples of a Gaussian filter, defaults to None
    :type n: int, optional
    :return: relative number of changes in tilt
    :rtype: int
    """
    return WritingNumberOfChangesUtils(sample_wrapper, fs=fs, fc=fc, n=n).get_relative_number_of_changes_in_tilt()


def relative_number_of_changes_in_pressure(sample_wrapper, fs, fc=None, n=None):
    """
    Returns relative number of changes in pressure.

    :param sample_wrapper: sample wrapper object
    :type sample_wrapper: HandwritingSampleWrapper
    :param fs: sampling frequency
    :type fs: float
    :param fc: cutoff frequency for the low-pass filter, defaults to None
    :type fc: float, optional
    :param n: number of samples of a Gaussian filter, defaults to None
    :type n: int, optional
    :return: relative number of changes in pressure
    :rtype: int
    """
    return WritingNumberOfChangesUtils(sample_wrapper, fs=fs, fc=fc, n=n).get_relative_number_of_changes_in_pressure()


def relative_number_of_changes_in_velocity_profile(sample_wrapper, fs, fc=None, n=None):
    """
    Returns relative number of changes in velocity profile.

    :param sample_wrapper: sample wrapper object
    :type sample_wrapper: HandwritingSampleWrapper
    :param fs: sampling frequency
    :type fs: float
    :param fc: cutoff frequency for the low-pass filter, defaults to None
    :type fc: float, optional
    :param n: number of samples of a Gaussian filter, defaults to None
    :type n: int, optional
    :return: relative number of changes in velocity profile
    :rtype: int
    """
    return WritingNumberOfChangesUtils(
        sample_wrapper=sample_wrapper,
        fs=fs,
        fc=fc,
        n=n).get_relative_number_of_changes_in_velocity_profile()



class FeatureNameMissingError(Exception):
    """Raised when a feature name is missing"""
    pass


class FeatureNameInvalidTypeError(Exception):
    """Raised when a feature name is of invalid type"""
    pass


class FeatureNameUnsupportedError(Exception):
    """Raised when a feature name us not supported"""
    pass


class FeatureArgumentMissingError(Exception):
    """Raised when a mandatory feature argument is missing"""
    pass


class FeatureArgumentInvalidTypeError(Exception):
    """Raised when a mandatory feature argument is of invalid type"""
    pass


class FeatureArgumentUnsupportedValueError(Exception):
    """Raised when a mandatory feature argument has unsupported value"""
    pass


class StatisticsForSingleValuedFeatureError(Exception):
    """Raised when statistics are to be computed for a single-valued feature"""
    pass



import numpy
from handwriting_features.features import HandwritingFeatures
from handwriting_features.features.configuration.mapping import HandwritingFeaturesMapping
from handwriting_features.features.validation import HandwritingFeaturesFusion, HandwritingFeaturesValidation
from handwriting_features.interface.featurizer.utils import (
    SingleSubjectFeatureUtils,
    MultiSubjectFeatureUtils,
    FeaturesPipelineUtils
)


class BaseFeatureExtractorHandler(object):
    """Base class for the feature extractor handlers"""

    # Features pipeline utils
    pipeline_utils = FeaturesPipelineUtils

    # Features
    features = HandwritingFeatures

    # Feature mapper
    mapping = HandwritingFeaturesMapping

    # Feature validator
    validation = HandwritingFeaturesValidation


class SingleSubjectFeatureExtractorHandler(BaseFeatureExtractorHandler):
    """Class implementing the single-subject features extractor handler"""

    # Feature utils
    utils = SingleSubjectFeatureUtils

    @classmethod
    def extract(cls, data_values, data_labels=None, pipeline=None, preparation=True, **configuration):
        """
        Extracts the features specified in the pipeline for a single subject.

        :param data_values: sample values to extract the features from
        :type data_values: numpy.ndarray
        :param data_labels: labels for data samples, defaults to None
        :type data_labels: list, optional
        :param pipeline: pipeline of the features, defaults to None
        :type pipeline: list, optional
        :param preparation: prepare the pipeline of features, defaults to True
        :type preparation: bool, optional
        :param configuration: common extractor configuration
        :type configuration: **kwargs
        :return: extracted features and labels
        :rtype: dict {"features": ..., "labels": ...}
        """

        # Initialize the handwriting features interface
        features = cls.features.from_numpy_array(data_values, data_labels, **configuration)

        # Initialize the feature mapping
        mapping = cls.mapping(features)

        # Prepare the buffers for extracted feature values/labels
        feature_values = []
        feature_labels = []

        # Prepare the features pipeline
        if preparation:
            pipeline = cls.pipeline_utils.prepare_features_pipeline(pipeline)

        # Extract the features specified in the features pipeline
        for feature in pipeline:

            # Get the feature name and args
            name = feature.get("name")
            args = feature.get("args", {})

            # Prepare the feature args (fuze the feature args with the common configuration)
            args = HandwritingFeaturesFusion.fuze(
                feature.get("name"),
                args,
                features.config,
                features.skip_features)

            # Get the validated feature arguments
            arguments = cls.validation.validate(name, args)

            # Extract the feature
            extracted = mapping.map(feature.get("name"))(**arguments)

            # Update the feature values/labels
            feature_values.append(cls.utils.prepare_feature_values(extracted))
            feature_labels.append(cls.utils.prepare_feature_labels(extracted, name, args))

        # Return the extracted feature values/labels
        return {
            "features": feature_values,
            "labels": feature_labels
        }


class MultiSubjectFeatureExtractorHandler(BaseFeatureExtractorHandler):
    """Class implementing the multi-subject features extractor handler"""

    # Feature extractor
    extractor = SingleSubjectFeatureExtractorHandler

    # Feature utils
    utils = MultiSubjectFeatureUtils

    @classmethod
    def extract(cls, data_values, data_labels=None, pipeline=None, **configuration):
        """
        Extracts the features specified in the pipeline for multiple subjects.

        :param data_values: samples values to extract the features from
        :type data_values: numpy.ndarray
        :param data_labels: labels for data samples, defaults to None
        :type data_labels: list, optional
        :param pipeline: pipeline of the features, defaults to None
        :type pipeline: list, optional
        :param configuration: common extractor configuration
        :type configuration: **kwargs
        :return: extracted features and labels
        :rtype: dict {"features": ..., "labels": ...}
        """

        # Prepare the features pipeline
        pipeline = cls.pipeline_utils.prepare_features_pipeline(pipeline)

        # Extract the features specified in the features pipeline for each subject
        extracted = [
            cls.extractor.extract(
                data_values=data_values[sample, ...],
                data_labels=data_labels,
                pipeline=pipeline,
                preparation=False,
                **configuration)
            for sample in range(data_values.shape[0])
        ]

        # Prepare the feature values/labels
        feature_values = cls.utils.prepare_feature_values(extracted, pipeline)
        feature_labels = cls.utils.prepare_feature_labels(extracted, pipeline)

        # Return the extracted feature values/labels
        return {
            "features": feature_values,
            "labels": feature_labels
        }



import numpy
from handwriting_features.data.utils.math import derivation
from handwriting_features.features.implementation.conventional.utils.spatial import IntersectionUtils, ProjectionUtils


def stroke_length(sample_wrapper, in_air):
    """
    Returns stroke length.

    :param sample_wrapper: sample wrapper object
    :type sample_wrapper: HandwritingSampleWrapper
    :param in_air: in-air flag
    :type in_air: bool
    :return: stroke length
    :rtype: numpy.ndarray or np.NaN
    """

    # Get the strokes
    strokes = sample_wrapper.on_surface_strokes \
        if not in_air \
        else sample_wrapper.in_air_strokes

    # Check the presence of strokes
    if not strokes:
        return numpy.nan

    # Get the strokes length
    length = [sum(numpy.sqrt(derivation(stroke.x) ** 2 + derivation(stroke.y) ** 2)) for stroke in strokes]
    length = numpy.array(length)

    # Return the length
    return length


def stroke_height(sample_wrapper, in_air):
    """
    Returns stroke height.

    :param sample_wrapper: sample wrapper object
    :type sample_wrapper: HandwritingSampleWrapper
    :param in_air: in-air flag
    :type in_air: bool
    :return: stroke height
    :rtype: numpy.ndarray or np.NaN
    """

    # Get the strokes
    strokes = sample_wrapper.on_surface_strokes \
        if not in_air \
        else sample_wrapper.in_air_strokes

    # Check the presence of strokes
    if not strokes:
        return numpy.nan

    # Return the stokes height
    return numpy.array([max(stroke.y) - min(stroke.y) for stroke in strokes])


def stroke_width(sample_wrapper, in_air):
    """
    Returns stroke width.

    :param sample_wrapper: sample wrapper object
    :type sample_wrapper: HandwritingSampleWrapper
    :param in_air: in-air flag
    :type in_air: bool
    :return: stroke width
    :rtype: numpy.ndarray or np.NaN
    """

    # Get the strokes
    strokes = sample_wrapper.on_surface_strokes \
        if not in_air \
        else sample_wrapper.in_air_strokes

    # Check the presence of strokes
    if not strokes:
        return numpy.nan

    # Return the stokes width
    return numpy.array([max(stroke.x) - min(stroke.x) for stroke in strokes])


def writing_length(sample_wrapper, in_air):
    """
    Returns writing length.

    :param sample_wrapper: sample wrapper object
    :type sample_wrapper: HandwritingSampleWrapper
    :param in_air: in-air flag
    :type in_air: bool
    :return: writing length
    :rtype: float
    """

    # Get the on-surface/in-air sample data
    sample = sample_wrapper.on_surface_data \
        if not in_air \
        else sample_wrapper.in_air_data

    # Check the presence of sample data
    if not sample:
        return numpy.nan

    # Return the writing length
    return sum(numpy.sqrt(derivation(sample.x) ** 2 + derivation(sample.y) ** 2))


def writing_height(sample_wrapper, in_air):
    """
    Returns writing height.

    :param sample_wrapper: sample wrapper object
    :type sample_wrapper: HandwritingSampleWrapper
    :param in_air: in-air flag
    :type in_air: bool
    :return: writing height
    :rtype: float
    """

    # Get the on-surface/in-air sample data
    sample = sample_wrapper.on_surface_data \
        if not in_air \
        else sample_wrapper.in_air_data

    # Check the presence of sample data
    if not sample:
        return numpy.nan

    # Return the writing height
    return float(numpy.max(sample.y) - numpy.min(sample.y))


def writing_width(sample_wrapper, in_air):
    """
    Returns writing width.

    :param sample_wrapper: sample wrapper object
    :type sample_wrapper: HandwritingSampleWrapper
    :param in_air: in-air flag
    :type in_air: bool
    :return: writing width
    :rtype: float
    """

    # Get the on-surface/in-air sample data
    sample = sample_wrapper.on_surface_data \
        if not in_air \
        else sample_wrapper.in_air_data

    # Check the presence of sample data
    if not sample:
        return numpy.nan

    # Return the writing width
    return float(numpy.max(sample.x) - numpy.min(sample.x))


def number_of_intra_stroke_intersections(sample_wrapper):
    """
    Returns number of intra-stroke intersections.

    :param sample_wrapper: sample wrapper object
    :type sample_wrapper: HandwritingSampleWrapper
    :return: number of intra-stroke intersections
    :rtype: numpy.ndarray or np.NaN
    """
    return IntersectionUtils(sample_wrapper).get_number_of_intra_stroke_intersections()


def relative_number_of_intra_stroke_intersections(sample_wrapper):
    """
    Returns relative number of intra-stroke intersections.

    :param sample_wrapper: sample wrapper object
    :type sample_wrapper: HandwritingSampleWrapper
    :return: relative number of intra-stroke intersections
    :rtype: numpy.ndarray or np.NaN
    """
    return IntersectionUtils(sample_wrapper).get_relative_number_of_intra_stroke_intersections()


def total_number_of_intra_stroke_intersections(sample_wrapper):
    """
    Returns total number of intra-stroke intersections.

    :param sample_wrapper: sample wrapper object
    :type sample_wrapper: HandwritingSampleWrapper
    :return: total number of intra-stroke intersections
    :rtype: int
    """
    return IntersectionUtils(sample_wrapper).get_total_number_of_intra_stroke_intersections()


def relative_total_number_of_intra_stroke_intersections(sample_wrapper):
    """
    Returns relative total number of intra-stroke intersections.

    :param sample_wrapper: sample wrapper object
    :type sample_wrapper: HandwritingSampleWrapper
    :return: relative total number of intra-stroke intersections
    :rtype: float
    """
    return IntersectionUtils(sample_wrapper).get_relative_total_number_of_intra_stroke_intersections()


def number_of_inter_stroke_intersections(sample_wrapper):
    """
    Returns number of inter-stroke intersections.

    :param sample_wrapper: sample wrapper object
    :type sample_wrapper: HandwritingSampleWrapper
    :return: number of inter-stroke intersections
    :rtype: int
    """
    return IntersectionUtils(sample_wrapper).get_number_of_inter_stroke_intersections()


def relative_number_of_inter_stroke_intersections(sample_wrapper):
    """
    Returns relative number of inter-stroke intersections.

    :param sample_wrapper: sample wrapper object
    :type sample_wrapper: HandwritingSampleWrapper
    :return: relative number of inter-stroke intersections
    :rtype: float
    """
    return IntersectionUtils(sample_wrapper).get_relative_number_of_inter_stroke_intersections()


def vertical_peaks_indices(sample_wrapper, fs, n=None):
    """
    Returns vertical peaks indices.

    :param sample_wrapper: sample wrapper object
    :type sample_wrapper: HandwritingSampleWrapper
    :param fs: sampling frequency
    :type fs: float
    :param n: number of samples of a Gaussian filter, defaults to None
    :type n: int, optional
    :return: vertical peaks indices
    :rtype: numpy.ndarray or np.NaN
    """
    return ProjectionUtils(sample_wrapper, fs=fs, n=n).vertical_peaks_indices()


def vertical_valleys_indices(sample_wrapper, fs, n=None):
    """
    Returns vertical valleys indices.

    :param sample_wrapper: sample wrapper object
    :type sample_wrapper: HandwritingSampleWrapper
    :param fs: sampling frequency
    :type fs: float
    :param n: number of samples of a Gaussian filter, defaults to None
    :type n: int, optional
    :return: vertical valleys indices
    :rtype: numpy.ndarray or np.NaN
    """
    return ProjectionUtils(sample_wrapper, fs=fs, n=n).vertical_valleys_indices()


def vertical_peaks_values(sample_wrapper, fs, n=None):
    """
    Returns vertical peaks values.

    :param sample_wrapper: sample wrapper object
    :type sample_wrapper: HandwritingSampleWrapper
    :param fs: sampling frequency
    :type fs: float
    :param n: number of samples of a Gaussian filter, defaults to None
    :type n: int, optional
    :return: vertical peaks values
    :rtype: numpy.ndarray or np.NaN
    """
    return ProjectionUtils(sample_wrapper, fs=fs, n=n).vertical_peaks_values()


def vertical_valleys_values(sample_wrapper, fs, n=None):
    """
    Returns vertical valleys values.

    :param sample_wrapper: sample wrapper object
    :type sample_wrapper: HandwritingSampleWrapper
    :param fs: sampling frequency
    :type fs: float
    :param n: number of samples of a Gaussian filter, defaults to None
    :type n: int, optional
    :return: vertical valleys values
    :rtype: numpy.ndarray or np.NaN
    """
    return ProjectionUtils(sample_wrapper, fs=fs, n=n).vertical_valleys_values()


def vertical_peaks_velocity(sample_wrapper, fs, n=None):
    """
    Returns vertical peaks velocity.

    :param sample_wrapper: sample wrapper object
    :type sample_wrapper: HandwritingSampleWrapper
    :param fs: sampling frequency
    :type fs: float
    :param n: number of samples of a Gaussian filter, defaults to None
    :type n: int, optional
    :return: vertical peaks velocity
    :rtype: numpy.ndarray or np.NaN
    """
    return ProjectionUtils(sample_wrapper, fs=fs, n=n).vertical_peaks_velocity()


def vertical_valleys_velocity(sample_wrapper, fs, n=None):
    """
    Returns vertical valleys velocity.

    :param sample_wrapper: sample wrapper object
    :type sample_wrapper: HandwritingSampleWrapper
    :param fs: sampling frequency
    :type fs: float
    :param n: number of samples of a Gaussian filter, defaults to None
    :type n: int, optional
    :return: vertical valleys velocity
    :rtype: numpy.ndarray or np.NaN
    """
    return ProjectionUtils(sample_wrapper, fs=fs, n=n).vertical_valleys_velocity()


def vertical_peaks_distance(sample_wrapper, fs, n=None):
    """
    Returns vertical peaks distance.

    :param sample_wrapper: sample wrapper object
    :type sample_wrapper: HandwritingSampleWrapper
    :param fs: sampling frequency
    :type fs: float
    :param n: number of samples of a Gaussian filter, defaults to None
    :type n: int, optional
    :return: vertical peaks distance
    :rtype: int
    """
    return ProjectionUtils(sample_wrapper, fs=fs, n=n).vertical_peaks_distance()


def vertical_valleys_distance(sample_wrapper, fs, n=None):
    """
    Returns vertical valleys distance.

    :param sample_wrapper: sample wrapper object
    :type sample_wrapper: HandwritingSampleWrapper
    :param fs: sampling frequency
    :type fs: float
    :param n: number of samples of a Gaussian filter, defaults to None
    :type n: int, optional
    :return: vertical valleys distance
    :rtype: int
    """
    return ProjectionUtils(sample_wrapper, fs=fs, n=n).vertical_valleys_distance()


def vertical_peaks_duration(sample_wrapper, fs, n=None):
    """
    Returns vertical peaks duration.

    :param sample_wrapper: sample wrapper object
    :type sample_wrapper: HandwritingSampleWrapper
    :param fs: sampling frequency
    :type fs: float
    :param n: number of samples of a Gaussian filter, defaults to None
    :type n: int, optional
    :return: vertical peaks duration
    :rtype:
    """
    return ProjectionUtils(sample_wrapper, fs=fs, n=n).vertical_peaks_duration()


def vertical_valleys_duration(sample_wrapper, fs, n=None):
    """
    Returns vertical valleys duration.

    :param sample_wrapper: sample wrapper object
    :type sample_wrapper: HandwritingSampleWrapper
    :param fs: sampling frequency
    :type fs: float
    :param n: number of samples of a Gaussian filter, defaults to None
    :type n: int, optional
    :return: vertical valleys duration
    :rtype:
    """
    return ProjectionUtils(sample_wrapper, fs=fs, n=n).vertical_valleys_duration()



class UnsupportedAxisError(Exception):
    """Raised when unsupported axis argument is used"""
    pass


class UnsupportedSurfaceMovementError(Exception):
    """Raised when unsupported surface movement is used"""
    pass
